/*
 * File:   AccountUtils.cpp
 * Author: ryan
 *
 * Created on November 19, 2011, 11:29 AM
 */

#include <QSettings>
#include <QSqlQuery>
#include <QSqlError>

#include "AccountUtils.h"
#include "Account.h"
#include "QHacc.h"
#include "AccountMapper.h"
#include "Logger.h"

AccountUtils::AccountUtils( QSqlDatabase& c, QHacc * eng ) {
  conn = &c;
  engine = eng;
}

AccountUtils::AccountUtils( ) {
}

AccountUtils::AccountUtils( const AccountUtils& orig ) {
  conn = orig.conn;
  engine = orig.engine;
}

QString AccountUtils::longName( const Account& a ) const {
  AccountMapper& mapper = engine->getAccountMapper( );
  QString name = a.getName( );
  QSettings settings;
  const QString SEP = settings.value( AccountSeparator ).toString( );

  int pid = a.getParentId( );
  while ( 0 != pid ) {
    Account pa = mapper.getById( pid );
    name = pa.getName( ) + SEP + name;
    pid = pa.getParentId( );
  }

  return name;
}

QString AccountUtils::longName( int id ) const {
  return longName( engine->getAccountMapper( ).getById( id ) );
}

Money AccountUtils::balanceOn( const Account& acct, const QDate& d,
    AccountBalance type ) const {
  Money obal( acct.getBalance( Opening ) );
  if ( Opening == type ) {
    return obal;
  }

  QString q = "SELECT s.reconcilestateid, sum( s.amount ) FROM split s JOIN trans_split ts ON ts.splitid=s.id JOIN transentry t ON t.id=ts.transactionid JOIN transactiontype tt ON t.typeid=tt.id WHERE s.accountid=? AND t.date<? and tt.iscalculable=1 GROUP BY s.reconcilestateid";
  QSqlQuery bquery;
  bquery.setForwardOnly( true );
  bquery.prepare( q );
  bquery.bindValue( 0, acct.getId( ) );
  bquery.bindValue( 1, d );

  bquery.exec( );

  Money recbal;
  Money allbal;

  while ( bquery.next( ) ) {
    ReconcileState state =
        static_cast<ReconcileState>( bquery.value( 0 ).toInt( ) );
    Money val( bquery.value( 1 ).toInt( ) );
    if ( Yes == state ) {
      recbal += val;
    }
    allbal += val;
  }

  if( acct.isRightPlus() ){
    recbal.negate();
    allbal.negate();
  }

  Money bal = ( Reconciled == type ? recbal : allbal ) + obal;

  return bal;
}

AccountType AccountUtils::getAccountType( int aid ) const {
  return engine->getAccountMapper( ).getById( aid ).getType();
}

bool AccountUtils::isRightPlus( int aid ) const {
  return engine->getAccountMapper( ).getById( aid ).isRightPlus();
}

QList<Account> AccountUtils::getAccounts( AccountType t ){
  QList<Account> accts;
  AccountMapper& amap = engine->getAccountMapper( );

  QSqlQuery query;
  query.setForwardOnly( true );
  query.prepare( "SELECT id FROM account WHERE accounttypeid=?" );
  query.bindValue( 0, static_cast<int>( t ) );

  if( query.exec( ) ){
    while( query.next() ){
      accts.append( amap.getById( query.value( 0 ).toInt() ) ) ;
    }
  }
  else{
    Logger::error( QString( "error retrieving accounts of type: %1").arg( t ) );
    Logger::error( query.lastError( ).text( ) );
    throw MapperException( query.lastError( ).text( ) );
  }

  return accts;
}

QList<Account> AccountUtils::getMostPopularAccounts( AccountType t, int topx ) {
  QList<Account> accts;
  AccountMapper& amap = engine->getAccountMapper( );

  QSqlQuery query;
  query.setForwardOnly( true );
  query.prepare( "SELECT a.id, COUNT( s.accountid ) AS num FROM account a JOIN split s ON s.accountid=a.id WHERE a.accounttypeid=? GROUP BY a.id ORDER BY num DESC" );
  query.bindValue( 0, static_cast<int>( t ) );

  if( query.exec( ) ){
    while( query.next() && topx >= 0 ){
      accts.append( amap.getById( query.value( 0 ).toInt() ) ) ;
      topx--;
    }
  }
  else{
    Logger::error( QString( "error retrieving accounts of type: %1").arg( t ) );
    Logger::error( query.lastError( ).text( ) );
    throw MapperException( query.lastError( ).text( ) );
  }

  return accts;
}

QList<Account> AccountUtils::getMostPopularPairAccounts( int aid, AccountType type, int topx ) {
  QList<Account> list;
  QSqlQuery query;
  query.setForwardOnly( true );
  // get transactions in the correct journal having
  // at least one split in the given account
  QString q = "SELECT s2.accountid, COUNT( s2.accountid ) AS num FROM transentry t JOIN trans_split ts1 ON t.id=ts1.transactionid JOIN trans_split ts2 ON t.id=ts2.transactionid JOIN split s1 ON ts1.splitid=s1.id JOIN split s2 ON ts2.splitid=s2.id JOIN account a ON a.id=s2.accountid WHERE s1.accountid=? AND s2.accountid<>? AND a.accounttypeid=? GROUP BY s2.accountid ORDER BY num DESC";
  query.prepare( q );
  query.bindValue( 0, aid );
  query.bindValue( 1, aid );
  query.bindValue( 2, static_cast<int>( type ) );

  AccountMapper& amap = QHacc::getInstance()->getAccountMapper();
  if ( query.exec( ) ) {
    while( query.next() && topx >= 0 ){
      list << amap.getById( query.value( 0 ).toInt() );
      topx--;
    }
  }

  return list;
}
