#include "BudgetUtils.h"

#include "QHacc.h"
#include "Logger.h"
#include "Budget.h"
#include "AccountUtils.h"
#include "AccountMapper.h"

#include <utility>

BudgetUtils::BudgetUtils( QSqlDatabase& c, QHacc * eng ) :
  conn( &c ), engine( eng ) {}

BudgetUtils::BudgetUtils( const BudgetUtils& orig ) :
  conn( orig.conn ), engine( orig.engine ){}

bool BudgetUtils::isOverBudget( const Budget& b, const QDate& end ) const {
  AccountMapper& amap = engine->getAccountMapper();

  std::map<Account, Money> balances = budgetBreakdown( b, end );
  bool isover = false;

  for( const auto& p : b.getLimits() ){
    Account acct = amap.getById( p.first );
    const Money& limit = p.second.first;

    Money currbal = balances.at( acct );

    if( limit < currbal ){
      isover = true;
    }
  }

  if( b.getOtherLimit().isNonZero() ){
    Account acct;
    const Money& limit = b.getOtherLimit();
    Money currbal = balances.at( acct );

    if( limit < currbal ){
      isover = true;
    }
  }

  return isover;
}

std::map<Account, Money> BudgetUtils::budgetBreakdown( const Budget& b,
                                                       const QDate& end ) const {
  AccountUtils& autils = engine->getAccountUtils();
  AccountMapper& amap = engine->getAccountMapper();

  std::map<Account, Money> returner;
  QList<Account> viewedaccounts;

  QDate startdate( end );
  int year;
  int month;
  int day;
  startdate.getDate( &year, &month, &day );

  switch( b.getTime() ){
    case MonthToDate:
      startdate.setDate(year, month, 1 );
      break;
    case RunningMonth:
      startdate = startdate.addDays( -end.daysInMonth() );
      break;
    case YearToDate:
      startdate.setDate( year, 1, 1 );
      break;
    case RunningYear:
      startdate = startdate.addYears( -1 );
      break;
    default:
      Logger::error( "Unhandled budgeting timeframe" );
  }

  QList<Account> allExpenseAccounts = autils.getAccounts( Expense );
  for( const auto& p : b.getLimits() ){
    int aid = p.first;
    Account acct = amap.getById( aid );
    bool includesubs = p.second.second;

    QList<Account> accounts;
    if( includesubs ){
      accounts = amap.getChildren( aid );
    }
    accounts.push_back( acct );

    Money currbal;
    Money startbal;
    for( const auto& acct : accounts ){
      currbal += autils.balanceOn( acct, end );
      startbal += autils.balanceOn( acct, startdate );
      viewedaccounts.push_back( acct );
      allExpenseAccounts.removeAll( acct );
    }

    returner.insert( std::make_pair( acct, currbal - startbal ) );
  }

  // if we have an "other" limit, check all expenses against accounts we
  // haven't viewed yet
  if( b.getOtherLimit().isNonZero() ){
    Money currbal;
    Money startbal;
    for( const Account& acct : allExpenseAccounts ){
      currbal += autils.balanceOn( acct, end );
      startbal += autils.balanceOn( acct, startdate );
    }

    returner.insert( std::make_pair( Account(), currbal - startbal ) );
  }

  return returner;
}
