/************************* * * * * * * * * * * * * ***************************
    Copyright (c) 1999-2010 Ryan Bobko
                       ryan@ostrich-emulators.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 ************************** * * * * * * * * * * * * **************************
 * Created on July 12, 2009, 5:39 PM
 */

#include <QSettings>
#include <cmath>

#include "Money.h"
#include "Logger.h"
#include "Constants.h"

const int Money::ROLLOVER = 100;
const int Money::DECIMALS = 2;

Money::Money( int v ) {
  val = v;
}

Money::Money( const Money& orig ) {
  val = orig.getValue( );
}

Money Money::operator+( const Money& o ) const {
  return Money( o.getValue( ) + val );
}

Money Money::operator-( const Money& o ) const {
  return Money( val - o.getValue( ) );
}

Money& Money::operator+=( const Money& o ) {
  val += o.getValue( );
  return *this;
}

Money& Money::operator-=( const Money& o ) {
  val -= o.getValue( );
  return *this;
}

bool Money::operator!=( const Money& o ) const {
  return val != o.getValue( );
}

bool Money::operator==( const Money& o ) const {
  return val == o.getValue( );
}

bool Money::operator==( const int intval ) const {
  return val == intval;
}

bool Money::operator!=( const int intval ) const {
  return val != intval;
}

bool Money::operator>( const Money& o ) const {
  return val > o.getValue( );
}

bool Money::operator<( const Money& o ) const {
  return val < o.getValue( );
}

bool Money::operator>( int i ) const {
  return val > i;
}

bool Money::operator<( int i ) const {
  return val < i;
}

Money& Money::operator=( const Money& o ) {
  if( this != &o ){
    val = o.getValue( );
  }
  return *this;
}

Money& Money::operator=( const int intval ) {
  val = intval;
  return *this;
}

Money& Money::operator=( const QString& s ) {
  val = Money::getMoney( s ).getValue( );
  return *this;
}

Money Money::operator-( ) const {
  return Money( -getValue( ) );
}

double Money::operator/( const Money& s ) const{
  return static_cast<double>( getValue() ) / static_cast<double>( s.getValue() );
}

Money Money::getMoney( const QString& st ) {
  // convert the string to dollars and cents

  // if we hit something that can't be converted using toInt(), we want to
  // return 0 to the calling function
  bool ok = true;

  QSettings settings;
  const QString CURRSEP = settings.value( CurrencySeparator ).toString( );


  // we only want to deal with positive values here; we'll add the minus later
  bool neg = false;
  QString s;
  if ( st.startsWith( "-" ) ) {
    neg = true;
    s = st.mid( 1 );
  }
  else if ( st.startsWith( CURRSEP ) ) s = "0" + st;
  else s = st;

  int i = s.indexOf( CURRSEP );
  int ret = s.left( i ).toInt( &ok )*(int) pow( 10.0, DECIMALS );

  if ( i != -1 ) { // we have a decimal point to worry about
    // only look at the most significant DECIMALS places
    QString cents = s.mid( ++i, DECIMALS );
    int l = cents.length( );

    bool tok = ok;
    for ( i = 1; i <= l; i++ ) {
      ret += cents.mid( i - 1, 1 ).toInt( &tok )*(int) pow( 10.0, DECIMALS - i );
      ok = ( ok && tok );
    }
  }

  if ( !ok ) ret = 0;

  if ( neg ) ret = 0 - ret;

  Money m;
  m.setValue( ret );
  return m;
}

QString Money::value( const QString& val ) {
  // get val's decimal places cleaned to DECIMALS
  QSettings settings;
  const QString CURRSEP = settings.value( CurrencySeparator ).toString( );

  QString ret = val;
  if ( val.contains( CURRSEP ) ) {
    // we have a currency separator; make sure the DECIMALS are right
    int cur = val.indexOf( CURRSEP ) + 1; // include the separator
    int decsHave = val.length( ) - cur;
    int decsNeeded = DECIMALS - decsHave;
    if ( decsNeeded > 0 ) { // add DECIMALS to fill out the value
      for ( int i = 0; i < decsNeeded; i++ ) {
        ret += "0";
      }
    }
    else if ( decsNeeded < 0 ) { // remove decimal places from the end
      ret = val.mid( 0, val.length( ) - decsNeeded );
    }
    // else we're already formatted correctly
  }
  else {
    // don't have a separator, return the string with the appropriate
    // number of decimal places added
    ret += CURRSEP;
    for ( int i = 0; i < DECIMALS; i++ ) {
      ret += "0";
    }
  }

  return ret;
}

QString Money::value( const Money& m ) {
  // convert the cents to a string of dollars and cents

  QSettings settings;
  const QString CURRSEP = settings.value( CurrencySeparator ).toString( );
  int cents = m.getValue( );
  // we want only positive values for our comparisons; we'll add the neg later
  bool neg = false;
  if ( cents < 0 ) {
    neg = true;
    cents = 0 - cents;
  }

  // ROLLOVER will not be a double, GUARANTEED!
  const int MYROLL = (int) pow( 10.0, DECIMALS );
  int dollars = cents / MYROLL;
  int ncents = cents % MYROLL;

  //Logger::debug( QString( "from %3 (DECIMALS: %4), dollars is %1; cents is %2" ).
  //    arg( dollars ).arg( ncents ).arg( cents ).arg( DECIMALS ) );

  QString ret = QString::number( dollars ) + CURRSEP;
  int rolls = DECIMALS - 1;
  while ( rolls > 0 ) {
    if ( ncents < ( 10 * rolls-- ) ) ret.append( "0" );
  }
  ret.append( QString::number( ncents ) );

  // now that we have the string, put a minus sign if necessary
  if ( neg ) ret.prepend( "-" );

  return ret;
}

int Money::getValue( ) const {
  return val;
}

void Money::setValue( int i ) {
  val = i;
}

QString Money::toString( ) const {
  return value( *this );
}

QString Money::toPositiveString( ) const {
  return value( isPositive( ) ? val : -val );
}

bool Money::isPositive( ) const {
  return val > 0;
}

bool Money::isZero( ) const {
  return 0 == val;
}

bool Money::isNonZero() const {
  return 0 != val;
}

bool Money::isNonNegative( ) const {
  return val >= 0;
}

bool Money::isNegative( ) const{
  return getValue() < 0;
}

void Money::negate( ) {
  val = 0 - val;
}

Money Money::opposite( ) const {
  return Money( -val );
}

double Money::toDouble( ) const {
  return toString( ).toDouble( );
}

bool Money::oppositeSign( const Money& o ) const {
  return ( ( isPositive() && o.isNegative() ) ||
           ( isNegative() && o.isPositive() ) );
}

Money Money::abs() const {
  return ( isNegative() ? opposite() : *this );
}
