/************************* * * * * * * * * * * * * ***************************
    Copyright (c) 1999-2010 Ryan Bobko
                       ryan@ostrich-emulators.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 ************************** * * * * * * * * * * * * **************************
 * Created on July 5, 2010, 9:55 PM
 */

#include "SplitMapper.h"


#include <QSqlDatabase>
#include <QSqlQuery>
#include <QSqlError>
#include <QVariant>

#include "QHacc.h"
#include "Account.h"
#include "Journal.h"
#include "Split.h"
#include "Loan.h"
#include "TransactionUtils.h"
#include "TransactionMapper.h"
#include "TransactionSchedule.h"
#include "AccountMapper.h"
#include "Logger.h"
#include "EngineNotifier.h"

TransactionUtils::TransactionUtils( QSqlDatabase& c, QHacc * eng ) {
  conn = &c;
  engine = eng;
}

TransactionUtils::TransactionUtils( ) {
}

TransactionUtils::TransactionUtils( const TransactionUtils& orig ) {
  conn = orig.conn;
  engine = orig.engine;
}

QList<FullTransaction> TransactionUtils::getTransactions( const Account& a,
    const Journal& j, const QDate& end, const QDate& start ) throw ( MapperException ) {
  QList<FullTransaction> list;
  bool useJ = ( IDableEntity::NOID != j.getId( ) );
  bool useStart = ( start.isValid( ) );
  bool useEnd = ( end.isValid( ) );

  QSqlQuery query;
  query.setForwardOnly( true );
  // get transactions in the correct journal having
  // at least one split in the given account
  QString q = "SELECT DISTINCT ts.transactionid FROM account a JOIN split s ON a.id=s.accountid JOIN trans_split ts ON s.id=ts.splitid JOIN transentry t ON ts.transactionid=t.id WHERE s.accountid=?";

  if ( useJ ) {
    q += " AND t.journalid=?";
  }

  if ( useStart ) {
    q += " AND t.date>=?";
  }
  if ( useEnd ) {
    q += " AND t.date<=?";
  }

  query.prepare( q );
  query.bindValue( 0, (int) a.getId( ) );
  int pos = 1;

  if ( useJ ) {
    query.bindValue( pos++, (int) j.getId( ) );
  }
  if ( useStart ) {
    query.bindValue( pos++, start );
  }
  if ( useEnd ) {
    query.bindValue( pos++, end );
  }

  if ( query.exec( ) ) {
    TransactionMapper& tmapper = engine->getTransactionMapper( );
    while ( query.next( ) ) {
      // ...now get each transaction and the splits that make it up
      int tid = query.value( 0 ).toInt( );
      FullTransaction trans( tmapper.getById( tid ) );
      const Transaction& t = trans.getTransaction( );

      for( const Split& s : engine->getSplitMapper( ).getAll( t ) ) {
        trans.add( s );
      }

      list.append( trans );
    }
  }
  else {
    Logger::error( query.lastError( ).text( ) );
    throw MapperException( query.lastError( ).text( ) );
  }

  return list;
}

QList<FullTransaction> TransactionUtils::getTransactions( const QString& payee ){
  QList<FullTransaction> list;

  QSqlQuery query;
  query.setForwardOnly( true );
  QString q = "SELECT DISTINCT ts.transactionid FROM account a JOIN split s ON a.id=s.accountid JOIN trans_split ts ON s.id=ts.splitid JOIN transentry t ON ts.transactionid=t.id WHERE t.payee=?";
  query.prepare( q );
  query.bindValue( 0, payee );

  if ( query.exec( ) ) {
    TransactionMapper& tmapper = engine->getTransactionMapper( );
    while ( query.next( ) ) {
      // ...now get each transaction and the splits that make it up
      int tid = query.value( 0 ).toInt( );
      FullTransaction trans( tmapper.getById( tid ) );
      const Transaction& t = trans.getTransaction( );

      foreach( const Split& s, engine->getSplitMapper( ).getAll( t ) ) {
        trans.add( s );
      }

      list.append( trans );
    }
  }
  else {
    Logger::error( query.lastError( ).text( ) );
    throw MapperException( query.lastError( ).text( ) );
  }

  return list;
}

void TransactionUtils::save( FullTransaction& ft ) {
  TransactionMapper& tmap = engine->getTransactionMapper( );
  SplitMapper& smap = engine->getSplitMapper( );

  Transaction t( ft.getTransaction( ) );
  QList<Split> splits;
  tmap.save( t );

  for( Split s : simplifySplits( ft.getSplits( ) ) ) {
    smap.save( s );
    linkSplitToTrans( s, t );
    splits << s;
  }

  ft.reset( t, splits );

  AccountMapper& amap = engine->getAccountMapper( );

  for( const Split& s : splits ) {
    Account acct( amap.getById( s.getAccountId( ) ) );
    amap.updateBalances( acct );
  }

  engine->getNotifier( )->transaction( ft );
}

void TransactionUtils::update( const FullTransaction& oldy, FullTransaction& ft ) {
  TransactionMapper& tmap = engine->getTransactionMapper( );
  SplitMapper& smap = engine->getSplitMapper( );
  AccountMapper& amap = engine->getAccountMapper( );

  FullTransaction old( this->getTransaction( oldy.getTransaction().getId() ) );

  Transaction t( ft.getTransaction( ) );

  Logger::debug( QString( "ft update for trans: %1" ).arg( t.getId() ) );
  tmap.update( t );

  QList<Split> splits;
  QList<Split> oldsplits;
  for( const Split& s : old.getSplits( ) ) {
    if ( Split::NOID != s.getId( ) ) {
      oldsplits << s;
    }
  }

  for( Split s : simplifySplits( ft.getSplits( ) ) ) {
    if ( oldsplits.contains( s ) ) {
      smap.update( s );
      oldsplits.removeAll( s );
    }
    else {
      smap.save( s );
      linkSplitToTrans( s, t );
    }

    splits << s;

    Account acct( amap.getById( s.getAccountId( ) ) );
    Logger::debug( QString( "updating balance for acct: %1" ).arg( acct.getName( ) ) );
    QTime timer;
    timer.start( );
    amap.updateBalances( acct );
    Logger::debug( QString( "done updating balance for acct: %1 in %2ms" ).arg( acct.getName( ) ).arg( timer.elapsed( ) ) );
  }

  // remove any splits still in the old list

  for( const Split& oldsplit : oldsplits ) {
    smap.remove( oldsplit );
    Account acct( amap.getById( oldsplit.getAccountId( ) ) );

    Logger::debug( QString( "updating balance for acct: %1" ).arg( acct.getName( ) ) );
    QTime timer;
    timer.start( );
    amap.updateBalances( acct, false );
    Logger::debug( QString( "done updating balance for acct: %1 in %2ms" ).arg( acct.getName( ) ).arg( timer.elapsed( ) ) );
  }

  ft.reset( t, splits );
  engine->getNotifier( )->transaction( old, ft );
}

void TransactionUtils::linkSplitToTrans( const Split& s, const Transaction& t )
throw ( MapperException ) {
  QSqlQuery query;
  query.setForwardOnly( true );
  query.prepare( "INSERT INTO trans_split( transactionid, splitid ) VALUES ( ?, ? )" );
  query.bindValue( 0, t.getId( ) );
  query.bindValue( 1, s.getId( ) );

  if ( !query.exec( ) ) {
    Logger::error( query.lastError( ).text( ) );
    throw MapperException( query.lastError( ).text( ) );
  }
}

QList<Split> TransactionUtils::simplifySplits( QList<Split> splits ) const {
  QHash<int, Split> acctSplitLkp;

  for( const Split& s : splits ) {
    Logger::debug( QString( "predupes: %1" ).arg( s.toString( ) ) );
  }

  for( Split s : splits ) {
    if ( s.getValue( ).isZero( ) ) {
      // remove splits with no value
      continue;
    }

    int aid = s.getAccountId( );
    if ( acctSplitLkp.contains( aid ) ) {
      // Logger::debug( QString( "perdupes merging %1 with %2" ).arg( s.toString( ) ).arg( acctSplitLkp[aid].toString( ) ) );
      acctSplitLkp[aid].setValue( acctSplitLkp[aid].getValue( ) + s.getValue( ) );
    }
    else {
      acctSplitLkp.insert( aid, s );
      // Logger::debug( QString( "predupes adding %1" ).arg( s.toString( ) ) );
    }
  }

  for( const Split& s : acctSplitLkp.values( ) ) {
    Logger::debug( QString( "in dupes: %1" ).arg( s.toString( ) ) );
  }

  return acctSplitLkp.values( );
}

FullTransaction TransactionUtils::getTransaction( int tid ) throw (MapperException ) {
  TransactionMapper& tmap = engine->getTransactionMapper( );
  SplitMapper& smap = engine->getSplitMapper( );

  Transaction t( tmap.getById( tid ) );
  QList<Split> splits( smap.getAll( t ) );

  return FullTransaction( t, splits );
}

FullTransaction TransactionUtils::getTransaction( const TransactionSchedule& ts ) throw ( MapperException ) {
  return getTransaction( ts.transid );
}

void TransactionUtils::remove( const FullTransaction& ft ) {
  TransactionMapper& tmap = engine->getTransactionMapper( );
  SplitMapper& smap = engine->getSplitMapper( );
  tmap.remove( ft.getTransaction( ) );

  for( const Split& s : ft.getSplits( ) ) {
    smap.remove( s );
  }
}

QStringList TransactionUtils::getPayees( const Account& acct ){
  return getPayees( acct.getId() );
}

QStringList TransactionUtils::getPayees( int aid ){
  QStringList list;
  QSqlQuery query;
  query.setForwardOnly( true );
  // get transactions in the correct journal having
  // at least one split in the given account
  if( IDableEntity::NOID == aid ){
      QString q = "SELECT DISTINCT t.payee FROM transentry t";
      query.prepare( q );
  }
  else{
    QString q = "SELECT DISTINCT t.payee FROM transentry t JOIN trans_split ts ON t.id=ts.transactionid JOIN split s ON s.id=ts.splitid WHERE s.accountid=?";
    query.prepare( q );
    query.bindValue( 0, aid );
  }
  if ( query.exec( ) ) {
    while( query.next()){
      list<<query.value( 0 ).toString();
    }
  }

  return list;
}

QStringList TransactionUtils::getMostPopularPayees( int fromAid, int toAid, int topx ){
  QStringList list;
  QSqlQuery query;
  query.setForwardOnly( true );
  // get transactions in the correct journal having
  // at least one split in the given account
  if( IDableEntity::NOID == fromAid ){
      QString q = "SELECT t.payee, COUNT( t.payee ) AS num FROM transentry t GROUP BY t.payee ORDER BY num DESC";
      query.prepare( q );
  }
  else{
    QString q = "SELECT t.payee, COUNT( t.payee ) AS num FROM transentry t JOIN trans_split ts1 ON t.id=ts1.transactionid JOIN trans_split ts2 ON t.id=ts2.transactionid JOIN split s1 ON ts1.splitid=s1.id JOIN split s2 ON ts2.splitid=s2.id WHERE s1.accountid=? AND s2.accountid=? GROUP BY t.payee ORDER BY num DESC";
    query.prepare( q );
    query.bindValue( 0, fromAid );
    query.bindValue( 1, toAid );
  }

  if ( query.exec( ) ) {
    while( query.next() && topx > 0 ){
      list<<query.value( 0 ).toString();
      topx--;
    }
  }

  return list;
}


void TransactionUtils::execute( TransactionSchedule& ts ){
  TransactionMapper& tmap = engine->getTransactionMapper( );
  SplitMapper& smap = engine->getSplitMapper( );

  Transaction trans = tmap.getBySchedule( ts );
  QList<Split> splits = smap.getAll( trans );
  trans.setType( Normal );
  trans.setDate( ts.getNextRun() );

  FullTransaction ft( trans, splits );
  save( ft );

  step( ts );
}

void TransactionUtils::step( TransactionSchedule& ts ){
  TransactionMapper& tmap = engine->getTransactionMapper( );
  int year;
  int month;
  int day;
  QDate next = ts.getNextRun();
  next.getDate( &year, &month, &day );

  switch( ts.getType() ){
    case Monthly:
      next = next.addMonths( 1 );
      break;
    case BiMonthly:
      next = next.addDays( next.daysInMonth() / 2 );
      break;
    case BiWeekly:
      next = next.addDays( 14 );
      break;
    case LastDay:
      next = next.addMonths( 1 );
      year = next.year();
      month = next.month();
      day = next.daysInMonth();
      next.setDate( year, month, day );
      break;
    case Annually:
      next = next.addYears( 1 );
      break;
    case Daily:
      next = next.addDays( 1 );
      break;
    case Weekly:
      next = next.addDays( 7 );
      break;
    default:
      break;
  }

  ts.setNextRun( next, ts.getType() );
  tmap.update( ts );
}
