/************************* * * * * * * * * * * * * ***************************
    Copyright (c) 1999-2010 Ryan Bobko
                       ryan@ostrich-emulators.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 ************************** * * * * * * * * * * * * **************************/

#include <QSqlQuery>
#include <QSqlError>
#include <QVariant>

#include "AccountMapper.h"
#include "Logger.h"
#include "QHacc.h"
#include "EngineNotifier.h"

AccountMapper::AccountMapper( const QSqlDatabase& db, const QHacc * e )
: DBMapper<Account>( db, e ) {
}

AccountMapper::AccountMapper( const AccountMapper& orig ) : DBMapper<Account>( orig ) {
}

Account AccountMapper::getById( int id ) throw ( MapperException ) {
  if ( cache.contains( id ) ) {
    return Account( *( cache.object( id ) ) );
  }
  else {
    Account * acct = new Account( id );
    QSqlQuery query;
    query.setForwardOnly( true );
    query.prepare( "SELECT parentid, name, openingbalance, banknumber, notes, taxrate, accounttypeid FROM account WHERE id=?" );
    query.bindValue( 0, id );
    if ( query.exec( ) && query.next( ) ) {
      acct->setParentId( query.value( 0 ).toInt( ) );
      acct->setName( query.value( 1 ).toString( ) );
      Money obal = Money( query.value( 2 ).toInt( ) );
      acct->setBalance( obal, Opening );
      acct->setNumber( query.value( 3 ).toString( ) );
      acct->setNotes( query.value( 4 ).toString( ) );
      acct->setTaxRate( query.value( 5 ).toInt( ) );
      acct->setType( (AccountType) ( query.value( 6 ).toInt( ) ) );

      updateBalances( *acct, false );

      cache.insert( id, acct );
      return Account( *acct );
    }
    else {
      Logger::error( QString("error retrieving accountid: %1").arg( id ) );
      Logger::error( query.lastError( ).text( ) );
      throw MapperException( query.lastError( ).text( ) );
    }
  }
}

void AccountMapper::save( Account& saveme ) throw ( MapperException ) {
  QSqlQuery query;
  query.setForwardOnly( true );
  query.prepare( "INSERT INTO account( parentid, name, openingbalance, banknumber, notes, taxrate, accounttypeid ) VALUES ( ?, ?, ?, ?, ?, ?, ? )" );
  query.bindValue( 0, 0 == saveme.getParentId( ) ? QVariant( QVariant::Int ) :
      saveme.getParentId( ) );
  query.bindValue( 1, saveme.getName( ) );
  query.bindValue( 2, saveme.getBalance( Opening ).getValue( ) );
  query.bindValue( 3, saveme.getNumber( ) );
  query.bindValue( 4, saveme.getNotes( ) );
  query.bindValue( 5, saveme.getTaxRate( ) );
  query.bindValue( 6, (int) saveme.getType( ) );

  if ( query.exec( ) ) {
    saveme.setId( query.lastInsertId( ).toInt( ) );
    engine->getNotifier( )->account( saveme );
  }
  else {
    Logger::error( query.lastError( ).text( ) );
    throw MapperException( query.lastError( ).text( ) );
  }
}

void AccountMapper::update( Account& updateme ) throw ( MapperException ) {
  Account act( getById( updateme.getId( ) ) );
  cache.remove( updateme.getId( ) );

  QSqlQuery query;
  query.setForwardOnly( true );
  query.prepare( "UPDATE account SET parentid=?, name=?, openingbalance=?, banknumber=?, notes=?, taxrate=?, accounttypeid=? WHERE id=?" );
  query.bindValue( 0, 0 == updateme.getParentId( ) ? QVariant( QVariant::Int ) :
      updateme.getParentId( ) );
  query.bindValue( 1, updateme.getName( ) );
  query.bindValue( 2, updateme.getBalance( Opening ).getValue( ) );
  query.bindValue( 3, updateme.getNumber( ) );
  query.bindValue( 4, updateme.getNotes( ) );
  query.bindValue( 5, updateme.getTaxRate( ) );
  query.bindValue( 6, (int) updateme.getType( ) );
  query.bindValue( 7, updateme.getId( ) );

  if ( updateme.getBalance( Opening ) != act.getBalance( Opening ) ||
      updateme.getType( ) != act.getType( ) ) {
    updateBalances( updateme, false );
  }

  if ( query.exec( ) ) {
    engine->getNotifier( )->account( act, updateme );
  }
  else {
    Logger::error( query.lastError( ).text( ) );
    throw MapperException( query.lastError( ).text( ) );
  }
}

void AccountMapper::remove( const Account& removeme ) throw ( MapperException ) {
  cache.remove( removeme.getId( ) );

  QSqlQuery query;
  query.setForwardOnly( true );
  query.prepare( "DELETE FROM account WHERE id=?" );
  query.bindValue( 0, removeme.getId( ) );
  if ( query.exec( ) ) {
    engine->getNotifier( )->account( removeme.getId( ) );
  }
  else {
    Logger::error( query.lastError( ).text( ) );
    throw MapperException( query.lastError( ).text( ) );
  }
}

QList<Account> AccountMapper::getAll( ) throw ( MapperException ) {
  QList<Account> list;
  QSqlQuery query;
  query.setForwardOnly( true );
  query.prepare( "SELECT id FROM account" );

  if ( !query.exec( ) ) {
    Logger::error( "error retrieving all accounts" );
    Logger::error( query.lastError( ).text( ) );
    throw MapperException( query.lastError( ).text( ) );
  }

  while ( query.next( ) ) {
    int id = query.value( 0 ).toInt( );
    list.append( getById( id ) );
  }

  return list;
}

QList<Account> AccountMapper::getChildren( int pid ) throw ( MapperException ) {
  QList<Account> list;
  QSqlQuery query;
  query.setForwardOnly( true );
  if ( 0 == pid ) {
    query.prepare( "SELECT id FROM account WHERE parentid IS NULL" );
  }
  else {
    query.prepare( "SELECT id FROM account WHERE parentid=?" );
    query.bindValue( 0, pid );
  }

  if ( !query.exec( ) ) {
    Logger::error( QString("error retrieving account children for accountid: %1").arg( pid ) );
    Logger::error( query.lastError( ).text( ) );
    throw MapperException( query.lastError( ).text( ) );
  }

  while ( query.next( ) ) {
    int id = query.value( 0 ).toInt( );
    list.append( getById( id ) );
  }

  return list;
}

void AccountMapper::updateBalances( Account& acct,
                                    bool doSignal ) throw ( MapperException ) {
  Logger::debug( QString( "updating balance %2(%1)" ).arg( acct.getId() ).arg( acct.getName() ) );
  // current balance is opening balance plus all splits
  const Account old( acct );
  Money opening( old.getBalance( Opening ) );
  Money current;
  Money reconciled;

  QSqlQuery bquery;
  bquery.setForwardOnly( true );
  QString q = "SELECT s.reconcilestateid, sum( s.amount ) FROM split s JOIN trans_split ts ON ts.splitid=s.id JOIN transentry t ON t.id=ts.transactionid JOIN transactiontype tt ON t.typeid=tt.id WHERE s.accountid=? and tt.iscalculable=1 GROUP BY s.reconcilestateid";
  bquery.prepare( q );
  bquery.bindValue( 0, acct.getId( ) );
  bquery.exec( );
  while ( bquery.next( ) ) {
    ReconcileState state =
        static_cast<ReconcileState>( bquery.value( 0 ).toInt( ) );
    int dbval = bquery.value( 1 ).toInt( );

    current += dbval;

    if( Yes == state ){
      reconciled += dbval;
    }
  }

  if( acct.isRightPlus() ){
    current.negate();
    reconciled.negate();
  }

    // set the current balance
  acct.setBalance( current + opening, Current );
  acct.setBalance( reconciled + opening, Reconciled );

  // recache for subsequent queries
  cache.insert( acct.getId( ), new Account( acct ) );

  if ( doSignal ) {
    engine->getNotifier( )->account( old, acct );
  }
}
