#include "BudgetMapper.h"

#include "AccountMapper.h"
#include "Logger.h"
#include "QHacc.h"
#include "EngineNotifier.h"

#include <QSqlQuery>
#include <QSqlError>
#include <QVariant>

BudgetMapper::BudgetMapper( const QSqlDatabase& db, const QHacc * e )
: DBMapper<Budget>( db, e ) {
}

BudgetMapper::BudgetMapper( const BudgetMapper& orig ) : DBMapper<Budget>( orig ) {
}

Budget BudgetMapper::getById( int id ) throw ( MapperException ) {
  if ( cache.contains( id ) ) {
    return Budget( *( cache.object( id ) ) );
  }
  else {
    Budget * budget = new Budget( id );
    QSqlQuery query;
    query.setForwardOnly( true );
    query.prepare( "SELECT name, timetypeid, otherlimit FROM budget WHERE id=?" );
    query.bindValue( 0, id );
    if ( query.exec( ) && query.next( ) ) {
      budget->setName( query.value( 0 ).toString( ) );
      BudgetTime bt = static_cast<BudgetTime>( query.value( 1 ).toInt() );
      budget->setTime( bt );
      budget->setOtherLimit( Money( query.value( 2 ).toInt() ) );

      QSqlQuery query2;
      query2.setForwardOnly( true );
      query2.prepare( "SELECT accountid, blimit, includesubs FROM budget_limit WHERE budgetid=?" );
      query2.bindValue( 0, id );
      if ( query2.exec( ) ){
        while( query2.next() ){
            budget->addBudgetItem( query2.value( 0 ).toInt(),
                                   Money( query2.value( 1 ).toInt() ),
                                   query2.value( 2 ).toBool() );
        }
      }
      else {
        Logger::error( QString("error retrieving budget line id: %1").arg( id ) );
        Logger::error( query2.lastError( ).text( ) );
        throw MapperException( query2.lastError( ).text( ) );
      }

      cache.insert( id, budget );
      return Budget( *budget );
    }
    else {
      Logger::error( QString("error retrieving budget id: %1").arg( id ) );
      Logger::error( query.lastError( ).text( ) );
      throw MapperException( query.lastError( ).text( ) );
    }
  }
}

void BudgetMapper::save( Budget& saveme ) throw ( MapperException ) {
  QSqlQuery query;
  query.setForwardOnly( true );
  query.prepare( "INSERT INTO budget( name, timetypeid, otherlimit ) VALUES ( ?, ?, ? )" );
  query.bindValue( 0, saveme.getName() );
  query.bindValue( 1, static_cast<int>( saveme.getTime() ) );
  query.bindValue( 2, saveme.getOtherLimit().getValue() );

  if ( query.exec( ) ) {
    saveme.setId( query.lastInsertId( ).toInt( ) );

    QSqlQuery query2;
    query2.setForwardOnly( true );
    query2.prepare( "INSERT INTO budget_limit( budgetid, accountid, blimit, includesubs ) VALUES ( ?, ?, ?, ? )" );
    query2.bindValue( 0, saveme.getId() );

    for( const auto& p : saveme.getLimits() ){
      int aid =  p.first;
      std::pair<Money, bool> pp = p.second;
      const Money& m = pp.first;
      bool inc = pp.second;

      if( m.isNonZero() ){
        query2.bindValue( 1, aid );
        query2.bindValue( 2, m.getValue() );
        query2.bindValue( 3, ( inc ? 1 : 0 ) );

        if( !query2.exec() ){
          Logger::debug( query2.executedQuery() );
          Logger::error( query2.lastError( ).text( ) );
          throw MapperException( query2.lastError( ).text( ) );
        }
      }
    }
  }
  else {
    Logger::error( query.lastError( ).text( ) );
    throw MapperException( query.lastError( ).text( ) );
  }
}

void BudgetMapper::update( Budget& updateme ) throw ( MapperException ) {
  remove( updateme );
  save( updateme );
}

void BudgetMapper::remove( const Budget& removeme ) throw ( MapperException ) {
  cache.remove( removeme.getId( ) );

  QSqlQuery query;
  query.setForwardOnly( true );
  query.prepare( "DELETE FROM budget_limit WHERE budgetid=?" );
  query.bindValue( 0, removeme.getId( ) );

  if( query.exec() ){
    QSqlQuery query2;
    query2.setForwardOnly( true );
    query2.prepare( "DELETE FROM budget WHERE id=?" );
    query2.bindValue( 0, removeme.getId( ) );
    if ( !query2.exec( ) ) {
      Logger::error( query2.lastError( ).text( ) );
      throw MapperException( query2.lastError( ).text( ) );
    }
  }
  else {
    Logger::error( query.lastError( ).text( ) );
    throw MapperException( query.lastError( ).text( ) );
  }
}

QList<Budget> BudgetMapper::getAll( ) throw ( MapperException ) {
  QList<Budget> list;
  QSqlQuery query;
  query.setForwardOnly( true );
  query.prepare( "SELECT id FROM budget" );

  if ( !query.exec( ) ) {
    Logger::error( "error retrieving all budgets" );
    Logger::error( query.lastError( ).text( ) );
    throw MapperException( query.lastError( ).text( ) );
  }

  while ( query.next( ) ) {
    int id = query.value( 0 ).toInt( );
    list.append( getById( id ) );
  }

  return list;
}

QList<Budget> BudgetMapper::getAllFor( int aid ) throw ( MapperException) {
  QList<Budget> list;
  QSqlQuery query;
  query.setForwardOnly( true );
  query.prepare( "SELECT DISTINCT budgetid FROM budget_limit WHERE accountid=?" );
  query.bindValue( 0, aid );

  if ( !query.exec( ) ) {
    Logger::error( "error retrieving some budgets" );
    Logger::error( query.lastError( ).text( ) );
    throw MapperException( query.lastError( ).text( ) );
  }

  while ( query.next( ) ) {
    int id = query.value( 0 ).toInt( );
    list.append( getById( id ) );
  }

  return list;
}
