#include "LoanMapper.h"

#include <QSqlQuery>
#include <QSqlError>
#include <QVariant>

#include "Logger.h"
#include "QHacc.h"
#include "EngineNotifier.h"

LoanMapper::LoanMapper( const QSqlDatabase& db, const QHacc * e )
: DBMapper<Loan>( db, e ) {
}

LoanMapper::LoanMapper( const LoanMapper& orig ) : DBMapper<Loan>( orig ) {
}

Loan LoanMapper::getById( int id ) throw ( MapperException ) {
  if ( cache.contains( id ) ) {
    return Loan( *( cache.object( id ) ) );
  }
  else {
    Loan * acct = new Loan( id );
    QSqlQuery query;
    query.setForwardOnly( true );
    query.prepare( "SELECT principalid, interestid, periods, interest, payment, amount, name FROM loan WHERE id=?" );
    query.bindValue( 0, id );
    if ( query.exec( ) && query.next( ) ) {
      acct->setPrincipalAccountId( query.value( 0 ).toInt( ) );
      acct->setInterestAccountId( query.value( 1 ).toInt( ) );
      acct->setPeriods( query.value( 2 ).toInt( ) );
      acct->setYearlyInterestRate( query.value( 3 ).toDouble());
      acct->setPayment( Money( query.value( 4 ).toInt() ) );
      acct->setAmount( Money( query.value( 5 ).toInt() ) );
      acct->setName( query.value( 6 ).toString() );

      cache.insert( id, acct );
      return Loan( *acct );
    }
    else {
      Logger::error( QString("error retrieving loan id: %1").arg( id ) );
      Logger::error( query.lastError( ).text( ) );
      throw MapperException( query.lastError( ).text( ) );
    }
  }
}

void LoanMapper::save( Loan& saveme ) throw ( MapperException ) {
  QSqlQuery query;
  query.setForwardOnly( true );
  query.prepare( "INSERT INTO loan( principalid, interestid, periods, interest, payment, amount, name ) VALUES ( ?, ?, ?, ?, ?, ?, ? )" );
  query.bindValue( 0, saveme.getPrincipalAccountId() );
  query.bindValue( 1, saveme.getInterestAccountId() );
  query.bindValue( 2, saveme.getPeriods() );
  query.bindValue( 3, saveme.getYearlyInterestRate() );
  query.bindValue( 4, saveme.getPayment().getValue() );
  query.bindValue( 5, saveme.getAmount().getValue() );
  query.bindValue( 6, saveme.getName() );

  if ( query.exec( ) ) {
    saveme.setId( query.lastInsertId( ).toInt( ) );
  }
  else {
    Logger::error( query.lastError( ).text( ) );
    throw MapperException( query.lastError( ).text( ) );
  }
}

void LoanMapper::update( Loan& updateme ) throw ( MapperException ) {
  Loan act( getById( updateme.getId( ) ) );
  cache.remove( updateme.getId( ) );

  QSqlQuery query;
  query.setForwardOnly( true );
  query.prepare( "UPDATE loan SET principalid=?, interestid=?, periods=?, interest=?, payment=?, amount=?, name=? WHERE id=?" );
  query.bindValue( 0, updateme.getPrincipalAccountId() );
  query.bindValue( 1, updateme.getInterestAccountId() );
  query.bindValue( 2, updateme.getPeriods() );
  query.bindValue( 3, updateme.getYearlyInterestRate() );
  query.bindValue( 4, updateme.getPayment().getValue() );
  query.bindValue( 5, updateme.getAmount().getValue() );
  query.bindValue( 6, updateme.getName() );
  query.bindValue( 7, updateme.getId( ) );

  if ( !query.exec( ) ) {
    Logger::error( query.lastError( ).text( ) );
    throw MapperException( query.lastError( ).text( ) );
  }
}

void LoanMapper::remove( const Loan& removeme ) throw ( MapperException ) {
  cache.remove( removeme.getId( ) );

  QSqlQuery query;
  query.setForwardOnly( true );
  query.prepare( "DELETE FROM loan WHERE id=?" );
  query.bindValue( 0, removeme.getId( ) );
  if ( !query.exec( ) ) {
    Logger::error( query.lastError( ).text( ) );
    throw MapperException( query.lastError( ).text( ) );
  }
}

QList<Loan> LoanMapper::getAll( ) throw ( MapperException ) {
  QList<Loan> list;
  QSqlQuery query;
  query.setForwardOnly( true );
  query.prepare( "SELECT id FROM loan" );

  if ( !query.exec( ) ) {
    Logger::error( query.lastError( ).text( ) );
    throw MapperException( query.lastError( ).text( ) );
  }

  while ( query.next( ) ) {
    int id = query.value( 0 ).toInt( );
    list.append( getById( id ) );
  }

  return list;
}
