/************************* * * * * * * * * * * * * ***************************
    Copyright (c) 1999-2010 Ryan Bobko
                       ryan@ostrich-emulators.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 ************************** * * * * * * * * * * * * **************************/

#include "EngineNotifier.h"


#include <QSqlQuery>
#include <QSqlError>
#include <QVariant>

#include "SplitMapper.h"
#include "AccountMapper.h"
#include "Transaction.h"
#include "Logger.h"
#include "QHacc.h"

SplitMapper::SplitMapper( const QSqlDatabase& conn, const QHacc * e )
: DBMapper<Split>( conn, e ) {
  cache.setMaxCost( 2000 );
}

SplitMapper::SplitMapper( const SplitMapper& orig ) : DBMapper<Split>( orig ) {
}

Split SplitMapper::getById( int id ) throw ( MapperException ) {
  if ( cache.contains( id ) ) {
    return Split( *( cache.object( id ) ) );
  }
  else {
    Split * split = new Split( id );
    QSqlQuery query;
    query.setForwardOnly( true );
    query.prepare( "SELECT amount, reconcilestateid, reconciledate, taxrate, memo, accountid FROM split WHERE id=?" );
    query.bindValue( 0, QVariant( id ) );
    if ( query.exec( ) && query.next( ) ) {
      split->setValue( Money( query.value( 0 ).toInt( ) ) );
      split->setReconciled( static_cast<ReconcileState>( ( query.value( 1 ).toInt( ) ) ),
          query.value( 2 ).toDate( ) );
      split->setTaxRate( query.value( 3 ).toInt( ) );
      split->setMemo( query.value( 4 ).toString( ) );
      split->setAccountId( query.value( 5 ).toInt( ) );

      cache.insert( id, split );
      return Split( *split );
    }
    else {
      Logger::error( query.lastError( ).text( ) );
      throw MapperException( query.lastError( ).text( ) );
    }
  }
}

void SplitMapper::save( Split& t ) throw ( MapperException ) {
  QSqlQuery query;
  query.setForwardOnly( true );
  query.prepare( "INSERT INTO split( accountid, amount, reconcilestateid, reconciledate, taxrate, memo ) VALUES ( ?, ?, ?, ?, ?, ? )" );
  query.bindValue( 0, t.getAccountId( ) );
  query.bindValue( 1, t.getValue( ).getValue( ) );
  query.bindValue( 2, static_cast<int>( t.getReconciled( ) ) );
  query.bindValue( 3, t.getReconcileDate( ) );
  query.bindValue( 4, t.getTaxRate( ) );
  query.bindValue( 5, t.getMemo( ) );

  if ( query.exec( ) ) {
    t.setId( query.lastInsertId( ).toInt( ) );
    engine->getNotifier( )->split( t );
  }
  else {
    Logger::error( query.lastError( ).text( ) );
    throw MapperException( query.lastError( ).text( ) );
  }
}

void SplitMapper::update( Split & t ) throw ( MapperException ) {
  Split old( getById( t.getId( ) ) );

  if( old.getAccountId() == t.getAccountId() && old.getMemo() == t.getMemo()
      && old.getMemo() == t.getMemo() && old.getReconciled() == t.getReconciled()
      && old.getReconcileDate() == t.getReconcileDate()
      && old.getTaxRate() == t.getTaxRate() && old.getValue() == t.getValue() ){
    return;
  }


  cache.remove( t.getId( ) );
  QSqlQuery query;
  query.setForwardOnly( true );
  query.prepare( "UPDATE split set amount=?, reconcilestateid=?, reconciledate=?, taxrate=?, memo=?, accountid=? WHERE id=?" );
  query.bindValue( 0, t.getValue( ).getValue( ) );
  query.bindValue( 1, static_cast<int>( t.getReconciled( ) ) );
  query.bindValue( 2, t.getReconcileDate( ) );
  query.bindValue( 3, t.getTaxRate( ) );
  query.bindValue( 4, t.getMemo( ) );
  query.bindValue( 5, t.getAccountId( ) );
  query.bindValue( 6, t.getId( ) );

  if ( query.exec( ) ) {
    engine->getNotifier( )->split( old, t );
  }
  else {
    throw new MapperException( query.lastError( ).text( ) );
  }
}

void SplitMapper::remove( const Split& t ) throw ( MapperException ) {
  remove( t.getId( ) );
}

void SplitMapper::remove( int id ) throw ( MapperException ) {
  cache.remove( id );

  QSqlQuery query;
  query.setForwardOnly( true );
  query.prepare( "DELETE FROM split WHERE id=?" );
  query.bindValue( 0, id );

  if ( query.exec( ) ) {
    engine->getNotifier( )->split( id );
  }
  else {
    Logger::error( query.lastError( ).text( ) );
    throw MapperException( query.lastError( ).text( ) );
  }

  // SQLite doesn't do FK deletions for you
  query.prepare( "DELETE FROM trans_split WHERE splitid=?" );
  query.bindValue( 0, id );

  if ( query.exec( ) ) {
    engine->getNotifier( )->split( id );
  }
  else {
    Logger::error( query.lastError( ).text( ) );
    throw MapperException( query.lastError( ).text( ) );
  }

}

QList<Split> SplitMapper::getAll( ) throw ( MapperException ) {
  QList<Split> list;
  QSqlQuery query;
  query.setForwardOnly( true );
  query.prepare( "SELECT id FROM split" );

  if ( !query.exec( ) ) {
    Logger::error( query.lastError( ).text( ) );
    throw MapperException( query.lastError( ).text( ) );
  }

  while ( query.next( ) ) {
    int id = query.value( 0 ).toInt( );
    list.append( getById( id ) );
  }

  return list;
}

QList<Split> SplitMapper::getAll( const Transaction& t ) throw ( MapperException ) {
  QList<Split> list;
  QSqlQuery query;
  query.setForwardOnly( true );
  query.prepare( "SELECT splitid FROM trans_split WHERE transactionid=?" );
  query.bindValue( 0, t.getId( ) );

  if ( !query.exec( ) ) {
    Logger::error( QString("error retrieving splits for transid: %1").arg( t.getId() ) );
    Logger::error( query.lastError( ).text( ) );
    throw MapperException( query.lastError( ).text( ) );
  }
  Logger::debug( query.executedQuery() );

  while ( query.next( ) ) {
    int id = query.value( 0 ).toInt( );
    list.append( getById( id ) );
  }

  return list;
}

QList<Split> SplitMapper::getAll( const Account& a, const Journal& j,
    const QDate& end, const QDate& start ) throw ( MapperException ) {
  QList<Split> list;
  bool useJ = ( IDableEntity::NOID != j.getId( ) );
  bool useStart = ( start.isValid( ) );
  bool useEnd = ( end.isValid( ) );

  QSqlQuery query;
  query.setForwardOnly( true );
  // get transactions in the correct journal having
  // at least one split in the given account
  QString q = "SELECT s.id FROM split s JOIN trans_split ts ON s.id=ts.splitid JOIN transentry t ON ts.transactionid=t.id WHERE s.accountid=?";

  if ( useJ ) {
    q += " AND t.journalid=?";
  }

  if ( useStart ) {
    q += " AND t.date>=?";
  }
  if ( useEnd ) {
    q += " AND t.date<=?";
  }

  Logger::debug( q );

  query.prepare( q );
  query.bindValue( 0, (int) a.getId( ) );
  int pos = 1;

  if ( useJ ) {
    query.bindValue( pos++, (int) j.getId( ) );
  }
  if ( useStart ) {
    query.bindValue( pos++, start );
  }
  if ( useEnd ) {
    query.bindValue( pos++, end );
  }

  if ( query.exec( ) ) {
    SplitMapper& splits = engine->getSplitMapper( );
    while ( query.next( ) ) {
      int sid = query.value( 0 ).toInt( );
      list.append( splits.getById( sid ) );
    }
  }
  else {
    Logger::error( QString("error retrieving splits for accountid: %1").arg( a.getId() ) );
    Logger::error( query.lastError( ).text( ) );
    throw MapperException( query.lastError( ).text( ) );
  }

  return list;
}
