/************************* * * * * * * * * * * * * ***************************
    Copyright (c) 1999-2010 Ryan Bobko
                       ryan@ostrich-emulators.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 ************************** * * * * * * * * * * * * **************************/

#include "EngineNotifier.h"

#include "QHacc.h"

#include <QSqlQuery>
#include <QSqlError>
#include <QVariant>

#include "TransactionMapper.h"
#include "JournalMapper.h"
#include "Transaction.h"
#include "Logger.h"

TransactionMapper::TransactionMapper( const QSqlDatabase& conn, const QHacc * e )
  : DBMapper<Transaction>( conn, e ) {
}

TransactionMapper::TransactionMapper( const TransactionMapper& orig ) :
  DBMapper<Transaction>( orig ) {
}

Transaction TransactionMapper::getById( int id ) throw ( MapperException ) {
  if ( cache.contains( id ) ) {
    return Transaction( *( cache.object( id ) ) );
  }
  else {
    Transaction * transaction = new Transaction( id );
    QSqlQuery query;
    query.setForwardOnly( true );
    query.prepare( "SELECT typeid, num, date, payee, journalid FROM transentry WHERE id=?" );
    query.bindValue( 0, QVariant( id ) );

    if ( query.exec( ) && query.next( ) ) {
      Logger::debug( query.executedQuery() );

      transaction->setType( static_cast<TransactionType>( query.value( 0 ).toInt( ) ) );
      transaction->setNumber( query.value( 1 ).toString( ) );
      transaction->setDate( query.value( 2 ).toDate( ) );
      transaction->setPayee( query.value( 3 ).toString( ) );
      transaction->setJournalId( query.value( 4 ).toInt( ) );

      cache.insert( id, transaction );
      return Transaction( *transaction );
    }
    else {
      Logger::error( QString("error retrieving transid: %1").arg( id ) );
      Logger::error( query.lastError( ).text( ) );
      throw MapperException( query.lastError( ).text( ) );
    }
  }

}

Transaction TransactionMapper::getBySchedule( const TransactionSchedule& ts ) throw ( MapperException ) {
  return getById( ts.transid );
}

void TransactionMapper::save( Transaction& t ) throw ( MapperException ) {
  QSqlQuery query;
  query.setForwardOnly( true );
  query.prepare( "INSERT INTO transentry( typeid, num, date, payee, journalid ) VALUES ( ?, ?, ?, ?, ? )" );
  query.bindValue( 0, (int) t.getType( ) );
  query.bindValue( 1, t.getNumber( ) );
  query.bindValue( 2, t.getDate( ) );
  query.bindValue( 3, t.getPayee( ) );
  query.bindValue( 4, t.getJournalId( ) );

  if ( query.exec( ) ) {
    t.setId( query.lastInsertId( ).toInt( ) );
    engine->getNotifier( )->transaction( t );
  }
  else {
    Logger::error( query.lastError( ).text( ) );
    throw MapperException( query.lastError( ).text( ) );
  }
}

void TransactionMapper::update( Transaction & t ) throw ( MapperException ) {
  Transaction old( getById( t.getId( ) ) );
  cache.remove( t.getId( ) );

  Logger::debug( QString( "updating trans: %1, %2" ).arg( t.getId( ) ).arg( t.getPayee( ) ) );

  QSqlQuery query;
  query.setForwardOnly( true );
  query.prepare( "UPDATE transentry set typeid=?, num=?, date=?, payee=?, journalid=? WHERE id=?" );
  query.bindValue( 0, (int) t.getType( ) );
  query.bindValue( 1, t.getNumber( ) );
  query.bindValue( 2, t.getDate( ) );
  query.bindValue( 3, t.getPayee( ) );
  query.bindValue( 4, t.getJournalId( ) );
  query.bindValue( 5, t.getId( ) );

  if ( query.exec( ) ) {
    engine->getNotifier( )->transaction( old, t );
  }
  else {
    Logger::error( query.lastError( ).text( ) );
    throw MapperException( query.lastError( ).text( ) );
  }
}

void TransactionMapper::remove( const Transaction& t ) throw ( MapperException ) {
  cache.remove( t.getId( ) );

  unschedule( t );

  QSqlQuery query;
  query.setForwardOnly( true );
  query.prepare( "DELETE FROM transentry WHERE id=?" );
  query.bindValue( 0, t.getId( ) );

  if ( query.exec( ) ) {
    engine->getNotifier( )->transaction( t.getId( ) );
  }
  else {
    Logger::error( query.lastError( ).text( ) );
    throw MapperException( query.lastError( ).text( ) );
  }
}

QList<Transaction> TransactionMapper::getAll( ) throw ( MapperException ) {
  QList<Transaction> list;
  QSqlQuery query;
  query.setForwardOnly( true );
  query.prepare( "SELECT id FROM transentry" );

  if ( !query.exec( ) ) {
    Logger::error( "error retrieving all transactions" );
    Logger::error( query.lastError( ).text( ) );
    throw MapperException( query.lastError( ).text( ) );
  }

  while ( query.next( ) ) {
    int id = query.value( 0 ).toInt( );
    list.append( getById( id ) );
  }

  return list;
}

bool TransactionMapper::isScheduled( int transid ) throw ( MapperException ) {
  return !getSchedule( transid ).isNull( );
}

QList<TransactionSchedule> TransactionMapper::getSchedules( const Transaction& t ) throw ( MapperException ) {
  QList<TransactionSchedule> handles;

  QSqlQuery query;
  query.setForwardOnly( true );
  QString q = "SELECT id, scheduletypeid, lastrun, name, transactionid FROM schedule WHERE transactionid=?";
  query.prepare( q );
  query.bindValue( 0, t.getId() );

  if ( query.exec( ) ) {
    while( query.next( ) ) {
      int id = query.value( 0 ).toInt( );
      ScheduleType st = static_cast<ScheduleType>( query.value( 1 ).toInt() );
      QDate lr = query.value( 2 ).toDate();
      QString n = query.value( 3 ).toString();
      int tid = query.value( 4 ).toInt();
      handles.append( TransactionSchedule( id, st, tid, n, lr ) );
    }
  }
  else {
    Logger::error( query.lastError( ).text( ) );
    throw MapperException( query.lastError( ).text( ) );
  }

  return handles;
}

TransactionSchedule TransactionMapper::getSchedule( int schedid ) throw ( MapperException ) {
  TransactionSchedule handle;

  QSqlQuery query;
  query.setForwardOnly( true );
  QString q = "SELECT transactionid, scheduletypeid, nextrun, name FROM schedule WHERE id=?";
  query.prepare( q );
  query.bindValue( 0, schedid );

  if ( query.exec( ) ) {
    if ( query.next( ) ) {
      int tid = query.value( 0 ).toInt( );
      ScheduleType st = static_cast<ScheduleType>( query.value( 1 ).toInt() );
      QDate nr = query.value( 2 ).toDate();
      QString n = query.value( 3 ).toString();
      handle = TransactionSchedule( schedid, st, tid, n, nr );
    }
  }
  else {
    Logger::error( query.lastError( ).text( ) );
    throw MapperException( query.lastError( ).text( ) );
  }

  return handle;
}

TransactionSchedule TransactionMapper::getSchedule( const QString& name ) throw ( MapperException ) {
  TransactionSchedule handle;

  QSqlQuery query;
  query.setForwardOnly( true );
  QString q = "SELECT id FROM schedule WHERE name=?";
  query.prepare( q );
  query.bindValue( 0, name );

  if ( query.exec( ) ) {
    if( query.next( ) ) {
      handle = getSchedule( query.value( 0 ).toInt( ) );
    }
  }
  else {
    Logger::error( query.lastError( ).text( ) );
    throw MapperException( query.lastError( ).text( ) );
  }

  return handle;
}

Transaction TransactionMapper::getByName( const QString& name ) throw ( MapperException ) {
  Transaction trans;
  QSqlQuery query;
  query.setForwardOnly( true );
  QString q = "SELECT transactionid FROM schedule WHERE name=?";
  query.prepare( q );
  query.bindValue( 0, name );

  if ( query.exec( ) ) {
    if ( query.next( ) ) {
      int tid = query.value( 0 ).toInt( );
      trans = getById( tid );
    }
  }
  else {
    Logger::error( query.lastError( ).text( ) );
    throw MapperException( query.lastError( ).text( ) );
  }

  return trans;
}

QList<TransactionSchedule> TransactionMapper::getSchedules( ) throw ( MapperException ) {
  QList<TransactionSchedule> list;

  QSqlQuery query;
  query.setForwardOnly( true );
  QString q = "SELECT id FROM schedule";
  query.prepare( q );

  if ( query.exec( ) ) {
    while ( query.next( ) ) {
      list.append( getSchedule( query.value( 0 ).toInt( ) ) );
    }
  }
  else {
    Logger::error( query.lastError( ).text( ) );
    throw MapperException( query.lastError( ).text( ) );
  }

  return list;
}

void TransactionMapper::update( TransactionSchedule& updateme ) throw ( MapperException ) {
  QSqlQuery query;
  query.setForwardOnly( true );
  query.prepare( "UPDATE schedule SET transactionid=?, scheduletypeid=?, nextrun=?, name=? WHERE id=?" );
  query.bindValue( 0, updateme.transid );
  query.bindValue( 1, static_cast<int>( updateme.getType() ) );
  query.bindValue( 2, updateme.getNextRun() );
  query.bindValue( 3, updateme.getName() );
  query.bindValue( 4, updateme.getId() );

  if ( !query.exec( ) ) {
    Logger::error( query.lastError( ).text( ) );
    throw MapperException( query.lastError( ).text( ) );
  }
}

void TransactionMapper::unschedule( const TransactionSchedule& removeme ) throw ( MapperException ) {
  if ( removeme.isNull( ) ) {
    return;
  }

  QSqlQuery query;
  query.setForwardOnly( true );
  query.prepare( "DELETE FROM schedule WHERE id=?" );
  query.bindValue( 0, removeme.getId() );
  if ( !query.exec( ) ) {
    Logger::error( query.lastError( ).text( ) );
    throw MapperException( query.lastError( ).text( ) );
  }
}

void TransactionMapper::unschedule( const Transaction& removeme ) throw ( MapperException ) {
  QSqlQuery query;
  query.setForwardOnly( true );
  query.prepare( "DELETE FROM schedule WHERE transactionid=?" );
  query.bindValue( 0, removeme.getId() );
  if ( !query.exec( ) ) {
    Logger::error( query.lastError( ).text( ) );
    throw MapperException( query.lastError( ).text( ) );
  }
}

void TransactionMapper::schedule( const Transaction& t, TransactionSchedule &s ) throw ( MapperException ) {
  QSqlQuery query;
  query.setForwardOnly( true );
  query.prepare( "INSERT INTO schedule( transactionid, nextrun, scheduletypeid, name) VALUES ( ?, ?, ?, ? )" );
  query.bindValue( 0, t.getId() );
  query.bindValue( 1, s.getNextRun() );
  query.bindValue( 2, static_cast<int>( s.getType() ) );
  query.bindValue( 3, s.getName() );

  if ( query.exec( ) ) {
    s.setId( query.lastInsertId( ).toInt( ) );
  }
  else {
    Logger::error( query.lastError( ).text( ) );
    throw MapperException( query.lastError( ).text( ) );
  }
}
