/************************* * * * * * * * * * * * * ***************************
    Copyright (c) 1999-2010 Ryan Bobko
                       ryan@ostrich-emulators.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 ************************** * * * * * * * * * * * * **************************/

#include "AccountMapper.h"

#include <QTreeWidgetItem>
#include <QSettings>
#include <QHeaderView>

#include "AccountChooser.h"
#include "Account.h"
#include "Logger.h"
#include "QHacc.h"
#include "GUIStore.h"
#include "EngineNotifier.h"

AccountChooser::AccountChooser( QWidget * p ) : QTreeWidget( p ) {
  connect( this, SIGNAL( itemClicked( QTreeWidgetItem *, int ) ),
      this, SLOT( accountChanged( QTreeWidgetItem *, int ) ) );
  setObjectName( "AccountChooser" ); // for the stylesheet
  EngineNotifier * notifier = QHacc::getInstance( )->getNotifier( );

  connect( notifier, SIGNAL( accountChanged( const Account&, const Account& ) ),
      this, SLOT( accountUpdated( const Account&, const Account& ) ) );
  connect( notifier, SIGNAL( accountAdded( const Account& ) ),
      this, SLOT( refresh( ) ) );
}

void AccountChooser::saveSettings( ) {
  QSettings settings;
  QByteArray qba = header( )->saveState( );
  settings.setValue( "AccountChooser/headers", qba );
}

void AccountChooser::restoreSettings( ) {
  QSettings settings;
  QByteArray qba = settings.value( "AccountChooser/headers" ).toByteArray( );
  header( )->restoreState( qba );
}

const Account& AccountChooser::getCurrent( ) const {
  return current;
}

void AccountChooser::accountChanged( QTreeWidgetItem * item, int ) {
  int id = item->data( 0, Qt::UserRole ).toInt( );
  AccountMapper& amap = QHacc::getInstance( )->getAccountMapper( );
  current = amap.getById( id );
  emit accountSelected( current );
}

void AccountChooser::selectAccount( const Account& a ) {
  setCurrentItem( iditemlkp[a.getId( )] );
  scrollToItem( iditemlkp[a.getId( )] );
  accountChanged( iditemlkp[a.getId( )], 0 );
}

void AccountChooser::setAccount( const Account& a ){
  selectAccount( a );
  itemClicked( iditemlkp[a.getId()], 0 );
}

void AccountChooser::accountUpdated( const Account&, const Account& newer ) {
  if ( current.getId( ) == newer.getId( ) ) {
    current = newer;
  }

  QTreeWidgetItem * item = iditemlkp[newer.getId( )];
  item->setText( 0, newer.getName( ) );

  Money bal( newer.getBalance( Current ) );
  item->setText( 1, Money::value( bal ) );
  item->setData( 1, Qt::UserRole, bal.getValue( ) );
}

void AccountChooser::refresh( ) {
  clear( );
  iditemlkp.clear( );

  QHash<int, QHaccAccountTreeWidgetItem *> lookup;
  QList<int> todo;
  todo.append( 0 );
  QHacc * qhacc = QHacc::getInstance( );

  while ( !( todo.isEmpty( ) ) ) {

    int pid = todo.takeFirst( );

    for( const Account& acct : qhacc->getAccountMapper( ).getChildren( pid ) ) {
      int aid = acct.getId( );
      if ( current.getId( ) == acct.getId( ) ) {
        current = acct;
      }

      QHaccAccountTreeWidgetItem * item = NULL;
      if ( 0 == acct.getParentId( ) ) {
        item = new QHaccAccountTreeWidgetItem( acct, this );
      }
      else {

        item = new QHaccAccountTreeWidgetItem( acct, lookup[pid] );
      }

      lookup[aid] = item;
      iditemlkp[aid] = item;
      todo << aid;
    }
  }

  expandAll( );
  scrollToItem( iditemlkp[current.getId( )] );
}

QHaccAccountTreeWidgetItem::QHaccAccountTreeWidgetItem( const Account& a,
    AccountChooser * p ) : QTreeWidgetItem( p ) {

  init( a );
}

QHaccAccountTreeWidgetItem::QHaccAccountTreeWidgetItem( const Account& a,
    QHaccAccountTreeWidgetItem * parent ) : QTreeWidgetItem( parent ) {

  init( a );
}

void QHaccAccountTreeWidgetItem::init( const Account& acct ) {

  setText( 0, acct.getName( ) );
  setData( 0, Qt::UserRole, acct.getId( ) );
  setText( 1, Money::value( acct.getBalance( Current ) ) );
  setData( 1, Qt::UserRole, acct.getBalance( Current ).getValue( ) );
}

bool QHaccAccountTreeWidgetItem::operator<(const QTreeWidgetItem& other ) const {
  int sorton = treeWidget( )->sortColumn( );
  if ( 0 == sorton ) {
    return text( sorton ).toUpper( ) < other.text( sorton ).toUpper( );
  }
  else {
    int myval = data( 1, Qt::UserRole ).toInt( );
    int otherval = other.data( 1, Qt::UserRole ).toInt( );

    return myval < otherval;
  }
}
