/************************* * * * * * * * * * * * * ***************************
    Copyright (c) 1999-2010 Ryan Bobko
                       ryan@ostrich-emulators.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 ************************** * * * * * * * * * * * * **************************/

#include <QSortFilterProxyModel>
#include <QMouseEvent>
#include <QMenu>
#include <QAction>
#include <QSignalMapper>
#include <QCursor>
#include <QHeaderView>
#include <QSettings>
#include <QHashIterator>

#include <memory>
#include <iostream>

#include "SplitMapper.h"
#include "AccountViewer.h"
#include "GUIStore.h"
#include "Logger.h"
#include "QHacc.h"
#include "AccountMapper.h"
#include "TransactionMapper.h"
#include "EngineNotifier.h"
#include "Account.h"
#include "TransactionUtils.h"
#include "AccountUtils.h"
#include "DateDelegate.h"
#include "Loan.h"
#include "LoanMapper.h"
#include "LoanUtils.h"
#include "TransactionTypeFilter.h"

const int AccountViewer::TRANSROLE = DATEROLE + 1;
const int AccountViewer::ISTRANSROLE = TRANSROLE + 1;

AccountViewer::AccountViewer( const Journal& j, QWidget * p ) : TransactionViewer( p ) {
  TransactionViewer::setJournal( j );

  setUniformRowHeights( true );
  setAlternatingRowColors( false );
  setItemsExpandable( false );
  setSortingEnabled( true );
  setRootIsDecorated( false );
  setObjectName( "AccountViewer" ); // for the stylesheet
  QStringList headers;
  headers << tr( "Date" ) << tr( "Number" ) << tr( "Payee/Account" )
          << tr( "Memo" ) << tr( "Credit" ) << tr( "Debit" )
          << tr( "Rec", "reconciled field in account viewer" );
  setHeaderLabels( headers );

  header()->setStretchLastSection( false );

  connect( this, SIGNAL( itemActivated( QTreeWidgetItem *, int ) ),
           SLOT( transactionChanged( QTreeWidgetItem *, int ) ) );
  connect( this, SIGNAL( itemClicked( QTreeWidgetItem *, int ) ),
           SLOT( transactionChanged( QTreeWidgetItem *, int ) ) );

  EngineNotifier * notifier = QHacc::getInstance( )->getNotifier( );
  connect( notifier, SIGNAL( transactionAdded( const FullTransaction& ) ),
           SLOT( addTrans( const FullTransaction& ) ) );
  connect( notifier,
           SIGNAL( transactionChanged( const FullTransaction&, const FullTransaction& ) ),
           SLOT( updateTrans( const FullTransaction&, const FullTransaction& ) ) );
  connect( notifier,
           SIGNAL( splitChanged( const Split&, const Split& ) ),
           SLOT( updateSplit( const Split&, const Split& ) ) );

  setItemDelegateForColumn( 0, new DateDelegate( ) );
  for( int i = 1; i < columnCount(); i++ ){
    setItemDelegateForColumn( i, new TransactionDelegateBase( ) );
  }
  setItemDelegateForColumn( 1, new TransactionDelegateBase() );
  QSettings settings;
  showAll = settings.value( ShowAllTrans, "false" ).toBool();

  filters.push_back( std::unique_ptr<TransactionFilter>( new TransactionTypeFilter( Normal ) ) );
}

void AccountViewer::showAllSplits( bool showit ){
  showAll = showit;
}

void AccountViewer::keyPressEvent( QKeyEvent *event ){
  TransactionViewer::keyPressEvent( event );
  if( Qt::Key_I == event->key() ){
    Transaction t;
    t.setDate( currentItem()->data( 0, DATEROLE ).toDate() );
    emit editTrans( FullTransaction( t ) );
  }
  else if( Qt::Key_E == event->key() ){
    int tid = currentItem()->data( 0, TRANSROLE ).toInt();
    QHacc * engine = QHacc::getInstance();
    emit editTrans( engine->getTransactionUtils().getTransaction( tid ) );
  }
}

bool AccountViewer::shouldIncludeTrans( const FullTransaction& ft ) {
  bool ok = TransactionViewer::shouldIncludeTrans( ft );
  if( !loading ){
    ok = ok && ft.hasSplitForAccount( account );
  }

  return ok;
}

void AccountViewer::updateSplit( const Split&, const Split& newer ) {
  if ( sidItemLkp.contains( newer.getId( ) ) ) {
    Logger::debug( QString( "aviewer updating split: %1" ).arg( newer.getId() ) );
    QTreeWidgetItem * split = sidItemLkp[newer.getId( )];

    split->setData( 3, Qt::DisplayRole, newer.getMemo( ) );
    split->setData( ( newer.getValue( ) < 0 ? 5 : 4 ),
                    Qt::DisplayRole, newer.getValue( ).toPositiveString( ) );
    ReconcileState rs = newer.getReconciled( );
    if ( Yes == rs ) {
      split->setData( 6, Qt::DisplayRole, tr( "R", "reconciled state" ) );
    }
    else if ( Maybe == rs ) {
      split->setData( 6, Qt::DisplayRole, tr( "C", "cleared state" ) );
    }
    else {
      split->setData( 6, Qt::DisplayRole, "" );
    }
  }
}

void AccountViewer::updateTrans( const FullTransaction& older,
                                 const FullTransaction& newer ) {
  Logger::debug( QString( "updating trans, tid: %1" ).arg( newer.getTransaction().getId() ) );
  if ( tidItemLkp.contains( newer.getTransaction( ).getId( ) ) ) {
    Logger::debug( QString( "trans exists...updating" ) );
    for( const Split& s : older.getSplits( ) ) {
      Logger::debug( QString( "updating trans, removing sid: %1" ).arg( s.getId() ) );
      sidItemLkp.remove( s.getId( ) );
    }

    QHashIterator<int, QTreeWidgetItem *> i( sidItemLkp );
    while( i.hasNext() ){
      i.next();
      Logger::debug( QString( "split in cache %1" ).arg( i.key() ) );
    }


    QTreeWidgetItem * item = tidItemLkp[newer.getTransaction().getId()];
    tidItemLkp.remove( newer.getTransaction( ).getId( ) );
    delete item;

    addTrans( newer );
  }
}

void AccountViewer::addTrans( const FullTransaction& ft ) {
  if ( shouldIncludeTrans( ft ) ) {
    QHacc * qhacc = QHacc::getInstance( );
    AccountMapper& amap = qhacc->getAccountMapper( );
    AccountUtils& autils = qhacc->getAccountUtils( );

    QTreeWidgetItem * item = new QTreeWidgetItem( this );
    const Transaction& t = ft.getTransaction( );
    tidItemLkp[t.getId( )] = item;

    item->setData( 0, Qt::DisplayRole, t.getDate( ) );
    item->setData( 0, TRANSROLE, t.getId( ) );
    item->setData( 0, DATEROLE, t.getDate( ) );

    item->setData( 1, Qt::DisplayRole, t.getNumber( ) );
    item->setData( 2, Qt::DisplayRole, t.getPayee( ) );
    for( int i = 0; i < columnCount(); i++ ){
      item->setData( i, ISTRANSROLE, true );
    }

    const QList<Split>& splits = ft.getSplits( );

    if( showAll ){
      for( const Split& s : splits ){
        addSplit( s, item );
      }
    }
    else {
      QTreeWidgetItem * split = addSplit( ft[account], item );

      if( splits.length() > 2){
        split->setData( 2, Qt::DisplayRole, tr( "Split", "split transaction" ) );
      }
      else{
        // need to get the other split's account
        for( const Split& spl : splits ){
          if( spl.getAccountId() != account.getId()){
            split->setData( 2, Qt::DisplayRole,
                            autils.longName( amap.getById( spl.getAccountId( ) ) ) );
          }
        }
      }
    }

    expandItem( item );
    setCurrentItem( item );
  }
}

QTreeWidgetItem * AccountViewer::addSplit( const Split& s,
                                           QTreeWidgetItem * transitem ){
  QTreeWidgetItem * split = new QTreeWidgetItem( transitem );
  sidItemLkp[s.getId( )] = split;
  const Money& splitbal = s.getValue();
  QHacc * qhacc = QHacc::getInstance( );
  AccountMapper& amap = qhacc->getAccountMapper( );
  AccountUtils& autils = qhacc->getAccountUtils( );

  split->setData( 0, TRANSROLE, transitem->data(0, TRANSROLE ) );
  split->setData( 0, DATEROLE, transitem->data( 0, DATEROLE ) );

  split->setData( 2, Qt::DisplayRole,
                  autils.longName( amap.getById( s.getAccountId( ) ) ) );

  split->setData( 3, Qt::DisplayRole, s.getMemo( ) );
  split->setData( ( splitbal < 0 ? 5 : 4 ),
                  Qt::DisplayRole, splitbal.toPositiveString( ) );
  ReconcileState rs = s.getReconciled( );
  if ( Yes == rs ) {
    split->setData( 6, Qt::DisplayRole, tr( "R", "reconciled state" ) );
  }
  else if ( Maybe == rs ) {
    split->setData( 6, Qt::DisplayRole, tr( "C", "cleared state" ) );
  }

  return split;
}

void AccountViewer::transactionChanged( QTreeWidgetItem * item, int ) {

  int tid = item->data( 0, TRANSROLE ).toInt( );
  TransactionUtils& tutils = QHacc::getInstance( )->getTransactionUtils( );
  GUIStore::getInstance( )->setTransaction( tutils.getTransaction( tid ) );
}

void AccountViewer::showMenu( const QPoint & pos ) {
  rightclickpos = pos;
  QMenu menu( this );
  QSignalMapper * signaler = new QSignalMapper( this );
  connect( signaler, SIGNAL( mapped( const QString& ) ),
           SLOT( insertMem( const QString& ) ) );

  TransactionMapper& tmap = QHacc::getInstance()->getTransactionMapper();
  TransactionUtils& tutils = QHacc::getInstance()->getTransactionUtils();

  QList<TransactionSchedule> handles;
  QList<TransactionSchedule> thandles = tmap.getSchedules( );
  for( const TransactionSchedule& ts : thandles ){
    Transaction t = tmap.getBySchedule( ts );
    FullTransaction ft = tutils.getTransaction( t.getId() );
    if( ft.hasSplitForAccount( account ) ){
      handles.append( ts );
    }
  }

  if ( !handles.isEmpty( ) ) {
    QMenu * mems = menu.addMenu( tr( "Memorized Transactions" ) );

    for( const TransactionSchedule& th : handles ) {
      QAction * action = new QAction( th.getName( ), this );
      mems->addAction( action );

      connect( action, SIGNAL( triggered( ) ), signaler, SLOT( map( ) ) );
      signaler->setMapping( action, th.getName( ) );
    }
  }

  QList<Loan> loans;
  QList<Loan> tloans = QHacc::getInstance( )->getLoanMapper().getAll();
  for( Loan loan : tloans ){
    if( !( loan.getPrincipalAccountId() == account.getId() ||
           loan.getInterestAccountId() == account.getId() ) ){
      loans.append( loan );
    }
  }

  if( !loans.isEmpty() ){
    QMenu * lmenu = menu.addMenu( tr( "Loan Payments" ) );

    QSignalMapper * loansignals = new QSignalMapper( this );
    connect( loansignals, SIGNAL( mapped( int ) ), SLOT( insertLoan( int ) ) );

    for( Loan loan : loans ){
      QAction * action = new QAction( loan.getName( ), this );
      lmenu->addAction( action );

      connect( action, SIGNAL( triggered( ) ), loansignals, SLOT( map( ) ) );
      loansignals->setMapping( action, loan.getId() );
    }
  }

  QTreeWidgetItem * itemUnder = itemAt( pos );
  if( NULL != itemUnder ){
    int tid = itemUnder->data( 0, TRANSROLE ).toInt();
    QAction * domemo = new QAction( tr( "Memorize This Transaction" ), this );
    menu.addAction( domemo );

    QSignalMapper * memmapper = new QSignalMapper( this );
    connect( memmapper, SIGNAL( mapped( int ) ), SLOT( memorize( int ) ) );

    connect( domemo, SIGNAL( triggered() ), memmapper, SLOT( map() ) );
    memmapper->setMapping( domemo, tid );
  }

  menu.addSeparator();
  QAction * docopy = new QAction( tr( "Copy" ), this );
  connect( docopy, SIGNAL( triggered() ), SLOT( copyTrans() ) );
  menu.addAction( docopy );
  if( !copytrans.isNull() ){
    QAction * dopaste = new QAction( tr( "Paste" ), this );
    connect( dopaste, SIGNAL( triggered() ), SLOT( pasteTrans() ) );
    menu.addAction( dopaste );
  }

  menu.exec( mapToGlobal( pos ) );
}

void AccountViewer::memorize( int tid ){
  TransactionUtils& tutils = QHacc::getInstance()->getTransactionUtils();
  TransactionMapper& tmap = QHacc::getInstance()->getTransactionMapper();

  Transaction trans = tmap.getById( tid );
  SplitMapper& smap = QHacc::getInstance( )->getSplitMapper( );
  QString name = QString( "%1 %2" ).arg( trans.getPayee() ).
      arg( QDateTime::currentDateTime( ).toString( "MM/dd/yyyy hh:mm:ss.zzz" ) );
  TransactionSchedule sched;
  sched.setName( name );

  QList<Split> splits;
  for( const Split& tmp : smap.getAll( trans ) ){
    Split s( tmp );
    s.setReconciled( No );
    splits << s;
  }

  trans.setType( Template );
  FullTransaction ft( trans, splits );
  tutils.save( ft );
  tmap.schedule( ft.getTransaction(), sched );
}

void AccountViewer::mouseReleaseEvent( QMouseEvent * event ) {
  if ( Qt::RightButton == event->button( ) ) {
    showMenu( event->pos() );
    event->accept();
  }
  else {
    QTreeView::mouseReleaseEvent( event );
  }
}

void AccountViewer::insertMem( const QString& namedt ) {
  QTreeWidgetItem * idx = itemAt( rightclickpos );

  QDate date( nullptr == idx ?
                // if we're over a valid index, use its date, else current date
                QDate::currentDate() :
                idx->data( 0, DATEROLE ).toDate( ) );

  TransactionMapper& tmap = QHacc::getInstance( )->getTransactionMapper( );
  TransactionUtils& tutils = QHacc::getInstance( )->getTransactionUtils( );
  SplitMapper& smap = QHacc::getInstance( )->getSplitMapper( );

  Transaction t( tmap.getByName( namedt ) );
  t.setJournalId( journal.getId( ) );
  t.setDate( date );
  t.setType( Normal );

  FullTransaction ft( t, smap.getAll( t ) );
  tutils.save( ft );
}

void AccountViewer::insertLoan( int loanid ){
  Logger::debug( QString("insert loan: %1").arg( loanid ) );
  const Loan& loan = QHacc::getInstance( )->getLoanMapper().getById( loanid );
  LoanUtils& lutils = QHacc::getInstance()->getLoanUtils();

  QTreeWidgetItem * idx = itemAt( rightclickpos );

  QDate date( NULL == idx ?
                // if we're over a valid index, use its date, else current date
                QDate::currentDate() :
                idx->data( 0, DATEROLE ).toDate( ) );

  Transaction t;
  t.setJournalId( journal.getId() );
  t.setPayee( tr( "Loan Payment" ) );
  t.setDate( date );

  lutils.addPayment( loan, account, t );
}

void AccountViewer::cleanup() {
  sidItemLkp.clear();
  tidItemLkp.clear();
}

void AccountViewer::copy( int tid ) {
  TransactionUtils& tutils = QHacc::getInstance( )->getTransactionUtils( );
  copytrans = tutils.getTransaction( tid );
  Split split = copytrans[account];
  split.setAccountId( Split::NOID );
  copytrans.update( split );
}

void AccountViewer::pasteTrans(){
  if( !copytrans.isNull() ){
    TransactionUtils& tutils = QHacc::getInstance( )->getTransactionUtils( );

    Transaction t( copytrans.getTransaction() );
    t.setJournalId( journal.getId( ) );
    t.setType( Normal );

    FullTransaction ft( t, copytrans.getSplits() );
    Split split = copytrans[Split::NOID];
    split.setAccountId( account.getId() );
    ft.update( split );

    tutils.save( ft );
  }
}

void AccountViewer::copyTrans() {
  QTreeWidgetItem * idx = itemAt( rightclickpos );
  if( nullptr != idx ){
    copy( idx->data( 0, TRANSROLE ).toInt( ) );
  }
}
