/*
 * File:   DataEntry.cpp
 * Author: ryan
 */

#include <QDate>
#include <QSettings>
#include <QCompleter>
#include <QMessageBox>

#include <map>

#include "SplitMapper.h"
#include "DataEntry.h"
#include "Account.h"
#include "Journal.h"
#include "Logger.h"
#include "QHacc.h"
#include "AccountMapper.h"
#include "AccountUtils.h"
#include "TransactionMapper.h"
#include "AccountDelegate.h"
#include "ReconcileDelegate.h"
#include "GUIStore.h"
#include "FullTransaction.h"
#include "TransactionUtils.h"
#include "AccountDialog.h"
#include "Constants.h"
#include "SplitDialog.h"

DataEntry::DataEntry( QWidget * p ) : QWidget( p ) {
  widget.setupUi( this );

  QSettings settings;
  widget.dateEdit->setDisplayFormat( settings.value( DateFormat ).toString( ) );

  connect( widget.showsplits, SIGNAL( clicked() ), SLOT( openSplitEditor() ) );

  setupAccounts();

  QCompleter * qc = new QCompleter( &pcompleterModel, this );
  qc->setCaseSensitivity( Qt::CaseInsensitive );
  pcompleter.reset( qc );
  widget.payee->setCompleter( pcompleter.get() );

  vcompleter.reset( new QCompleter( &vcompleterModel, this ) );
  widget.amount->setCompleter( vcompleter.get() );
  vcompleter->setCompletionMode( QCompleter::UnfilteredPopupCompletion );

  connect( pcompleter.get(), SIGNAL( activated( const QString& ) ),
           SLOT( chooseCompletion( const QString& ) ) );
  loading = false;
}

DataEntry::~DataEntry( ) {}

void DataEntry::changeJournal( const Journal& j ) {
  journalid = j.getId( );
}

ReconcileState DataEntry::getRec( Qt::CheckState s ){
  ReconcileState rs;
  switch( s ){
    case Qt::PartiallyChecked:
      rs = Maybe;
      break;
    case Qt::Checked:
      rs = Yes;
      break;
    default:
      rs = No;
  }

  return rs;
}

Qt::CheckState DataEntry::getState( ReconcileState rs ){
  Qt::CheckState st;
  switch( rs ){
    case Maybe:
      st = Qt::PartiallyChecked;
      break;
    case No:
      st = Qt::Unchecked;
      break;
    default:
      st = Qt::Checked;
  }

  return st;
}

void DataEntry::openSplitEditor() {
  std::unique_ptr<SplitDialog> sd( new SplitDialog( this ) );
  sd->setSplits( fulltrans.getSplits() );
  sd->exec();
  if( QDialog::Accepted != sd->result() ){
    return;
  }

  QList<Split> splits;
  for( const Split& s : sd->splits() ){
    if( !s.getValue().isZero() ){
      splits << s;
    }
  }

  fulltrans.setSplits( splits );
  const Account& A = GUIStore::getInstance( )->getAccount( );
  const int AID = A.getId();
  Money zero;
  widget.amount->setText( zero.toString() );
  for( const Split& s : splits ){
    if( AID == s.getAccountId() ){
      widget.amount->setText( s.getValue().toString() );
      widget.reccheck->setCheckState(getState( s.getReconciled() ) );
    }
  }

  if( fulltrans.isSplit() ){
    widget.accountbox->setEnabled( false );
    widget.amount->setEnabled( false );
    widget.reccheck->setEnabled( false );
  }
  else {
    widget.accountbox->setEnabled( true );
    widget.amount->setEnabled( true );
    widget.reccheck->setEnabled( true );
    int idx = widget.accountbox->findData( fulltrans.otherAccountId( A ) );
    widget.accountbox->setCurrentIndex( idx );
  }
}

void DataEntry::edit( const FullTransaction& t ) {
  fulltrans = t;
  Logger::debug( QString( "data entry on transaction: %1" ).
      arg( t.getTransaction( ).getId( ) ) );
  widget.buttonBox->button( QDialogButtonBox::Ok )->setText( tr( "Save Changes" ) );
  setTransData( t.getTransaction( ) );
  setSplitData( t.getSplits( ) );

  Account A = GUIStore::getInstance( )->getAccount( );
  if( fulltrans.hasSplitForAccount( A ) ){
    Split splito( fulltrans[A] );
    if( A.isRightPlus() ){
      splito.setValue( splito.getValue().opposite() );
    }
    widget.amount->setText( splito.getValue().toString() );
    widget.reccheck->setCheckState( getState( splito.getReconciled() ) );
  }
  else {
    widget.buttonBox->button( QDialogButtonBox::Ok )->setText( tr( "Create" ) );
  }

  if( fulltrans.isSplit() ) {
    widget.accountbox->setEnabled( false );
    widget.amount->setEnabled( false );
    widget.reccheck->setEnabled( false );
  }
  else {
    widget.accountbox->setEnabled( true );
    widget.amount->setEnabled( true );
    widget.reccheck->setEnabled( true );
    int idx = widget.accountbox->findData( fulltrans.otherAccountId( A ) );
    widget.accountbox->setCurrentIndex( idx );
  }

  TransactionUtils& tutils = QHacc::getInstance()->getTransactionUtils();
  QSettings settings;
  if( settings.value( ShowAllPayees ).toBool() ){
    A = Account();
  }

  pcompleterModel.setStringList( tutils.getPayees( A ) );

  widget.payee->selectAll( );
  widget.payee->setFocus( );
}

void DataEntry::editNew( ) {
  Logger::debug( QString( "data entry on new transaction" ) );
  widget.accountbox->setEnabled( true );
  widget.amount->setEnabled( true );
  widget.buttonBox->button( QDialogButtonBox::Ok )->setText( tr( "Create" ) );

  widget.accountbox->setEnabled( true );
  widget.amount->setEnabled( true );
  widget.reccheck->setEnabled( true );

  Transaction t;
  t.setDate( QDate::currentDate( ) );
  t.setJournalId( journalid );

  setTransData( t );
  widget.dateEdit->setFocus( );

  Split s2; // a split for the other account

  QList<Split> list;
  list << s2;
  setSplitData( list );

  fulltrans.reset( t, list );

  TransactionUtils& tutils = QHacc::getInstance()->getTransactionUtils();
  Account A;
  QSettings settings;
  if( !settings.value( ShowAllPayees ).toBool() ){
    A = GUIStore::getInstance()->getAccount();
  }
  pcompleterModel.setStringList( tutils.getPayees( A ) );
}

void DataEntry::chooseCompletion( const QString& str ){
  QSettings settings;
  bool showall = settings.value( ShowAllPayees, false ).toBool();

  TransactionUtils& tutils = QHacc::getInstance()->getTransactionUtils();
  QList<FullTransaction> possibles = tutils.getTransactions( str );
  QStringList amounts;
  const Account& A = GUIStore::getInstance( )->getAccount( );
  Money money;
  std::map<int,int> acounts; // aid, count
  for( const FullTransaction& ft : possibles ) {
    if( ft.hasSplitForAccount( A ) ){
      const Split& split = ft[A];
      money = split.getValue();
      if( A.isRightPlus() ){
        money.negate();
      }
      amounts << money.toString();

      int AID = ft.otherAccountId( A );
      if( IDableEntity::NOID != AID ){
        acounts[AID]++;
      }
    }
    else if( showall ){
      for( const Split& split : ft.getSplits() ){
        money = split.getValue();
        if( A.isRightPlus() ){
          money.negate();
        }
        amounts << money.toString();
        acounts[split.getAccountId()]++;
      }
    }
  }

  vcompleterModel.setStringList( amounts );
  if( !loading ) {
    widget.amount->setText( money.toString() );
    widget.amount->selectAll();
    widget.amount->setFocus();
    vcompleter->complete();

    // pick the account with the most transactions, and select that one
    int aid = IDableEntity::NOID;
    int maxcount = 0;
    for( auto it : acounts ){
      Logger::debug( QString( "Possible account: %1: %2 trans" ).
                     arg( it.first ).arg( it.second ) );
      if( it.second > maxcount ){
        aid = it.first;
        maxcount = it.second;
      }
    }

    int idx = widget.accountbox->findData( aid );
    widget.accountbox->setCurrentIndex( idx );
  }
}

void DataEntry::saveData( ) {
  QHacc * engine = QHacc::getInstance( );

  Transaction trans( fulltrans.getTransaction( ) );
  trans.setJournalId( journalid );
  trans.setDate( widget.dateEdit->date( ) );
  trans.setNumber( widget.number->text( ) );
  trans.setPayee( widget.payee->text( ) );
  trans.setMemo( widget.memo->text() );

  bool ok;
  QList<Split> splits = getSplitsForSave( ok );
  if( ok ){
    FullTransaction newer( trans, splits );

    try{
      if ( Transaction::NOID == trans.getId( ) ) {
        engine->save( newer );
      }
      else {
        engine->update( fulltrans, newer );
      }
    }
    catch( MapperException me ){
      ok = false;
    }
  }

  if( !ok ){
    QMessageBox msgBox;
    msgBox.setText( tr( "An error occurred while saving this transaction.") );
    msgBox.exec();
  }

  // widget.accounts->saveSettings( "DataEntry/header" );
}

QList<Split> DataEntry::getSplitsForSave( bool &ok ){
  ok = true;
  QList<Split> splits;
  Money value = widget.amount->getMoney();
  const Account& A = GUIStore::getInstance( )->getAccount( );
  if( A.isRightPlus() ){
    value.negate();
  }

  bool isSplit = !widget.accountbox->isEnabled();

  if( isSplit ){
    // if we have a split transaction, the splits are already part of
    // fulltrans, so just return those
    splits = fulltrans.getSplits();
  }
  else{
    // else rebuild the splits based on the data entry fields

    int idx = widget.accountbox->currentIndex();
    int otherAID = widget.accountbox->itemData( idx ).toInt();
    if( IDableEntity::NOID == otherAID ){
      AccountDialog * ad = new AccountDialog( Account(), this );
      ad->exec();
      otherAID = ad->getCurrent().getId();
      ok=( Account::NOID != otherAID );
    }

    Split split;
    ReconcileState oldrs = No;
    if( fulltrans.hasSplitForAccount( A ) ){
      const Split& old = fulltrans[A];
      split = old;
      oldrs = old.getReconciled();
    }
    split.setAccountId( A.getId() );
    split.setMemo( widget.memo->text() );
    split.setValue( value );
    ReconcileState newrs = getRec( widget.reccheck->checkState() );
    if( newrs != oldrs ){
      split.setReconciled( newrs );
    }

    Split other;
    if( fulltrans.hasSplitForAccount( otherAID ) ){
      const Split& old = fulltrans[otherAID];
      other = old;
    }
    other.setAccountId( otherAID );
    other.setMemo( widget.memo->text() );
    other.setValue( value.opposite() );

    splits << split << other;
  }

  return splits;
}

void DataEntry::setTransData( const Transaction& t ) {
  loading = true;
  widget.dateEdit->setDate( t.getDate( ) );
  widget.number->setText( t.getNumber( ) );
  widget.payee->setText( t.getPayee( ) );
  widget.memo->setText( t.getMemo() );
  chooseCompletion( widget.payee->text() );

  widget.accountbox->setCurrentIndex( widget.accountbox->findData( Account::NOID ) );
  widget.reccheck->setChecked( false );
  Money m;
  widget.amount->setText( m.toString() );
  loading = false;
}

void DataEntry::setSplitData( const QList<Split>& ) {
}

void DataEntry::setupAccounts(){
    AccountMapper& amap = QHacc::getInstance( )->getAccountMapper( );
    AccountUtils& autils = QHacc::getInstance( )->getAccountUtils( );

    QHash<QString, int> accounts;
    QStringList anames;

    for( const Account& acct : amap.getAll( ) ) {
      QString longname = autils.longName( acct );
      accounts[longname] = acct.getId( );
      anames << longname;
    }
    anames.sort( );

    widget.accountbox->addItem( tr( "New Account" ), IDableEntity::NOID );

    for( const QString& s : anames ) {
      widget.accountbox->addItem( s, accounts[s] );
    }
}

void DataEntry::keyPressEvent( QKeyEvent * event ){
  QWidget::keyPressEvent( event );
  if( Qt::Key_Escape == event->key() ){
    event->accept();
    setVisible( false );
  }
}
