/************************* * * * * * * * * * * * * ***************************
    Copyright (c) 1999-2010 Ryan Bobko
                       ryan@ostrich-emulators.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 ************************** * * * * * * * * * * * * **************************/

#include <QSettings>
#include <iostream>

#include "JournalViewer.h"
#include "JournalMapper.h"
#include "AccountViewer.h"
#include "Journal.h"
#include "QHacc.h"
#include "GUIStore.h"
#include "FullTransaction.h"
#include "DateRangeFilter.h"
#include "CompositeFilter.h"
#include "ReconcileFilter.h"

JournalViewer::JournalViewer( QWidget * p ) : QTabWidget( p ) {
  connect( this, SIGNAL( currentChanged( int ) ),
           SLOT( changeJournal( int ) ) );
}

void JournalViewer::refresh( ) {
  QHacc * qhacc = QHacc::getInstance( );

  disconnect( this, SIGNAL( currentChanged( int ) ),
              this, SLOT( changeJournal( int ) ) );
  bool first = true;

  for( const Journal& j : qhacc->getJournalMapper( ).getAll( ) ) {
    AccountViewer * viewer = new AccountViewer( j );
    connect( viewer, SIGNAL( editTrans( const FullTransaction& ) ),
             SLOT( sendEditTrans( const FullTransaction & ) ) );
    int tabidx = addTab( viewer, j.getName( ) );
    idJournalLkp[tabidx] = j;
    journalIdLkp[j.getId( )] = tabidx;

    if ( first ) {
      current = j;
      GUIStore::getInstance( )->setJournal( current );
      first = false;
    }
  }

  connect( this, SIGNAL( currentChanged( int ) ),
           SLOT( changeJournal( int ) ) );
  emit journalSelected( current );
}

void JournalViewer::changeJournal( int id ) {
  current = idJournalLkp[id];
  emit journalSelected( current );
}

void JournalViewer::changeFilters(){
  filtering = false;
  QSettings settings;
  int months = settings.value( FiltersTime, 0 ).toInt();
  const Account& account = GUIStore::getInstance()->getAccount();

  for ( int i = 0; i < count( ); i++ ) {
    AccountViewer * viewer = static_cast<AccountViewer *> ( widget( i ) );
    viewer->clearCustomFilters();

    CompositeFilter * compOr = nullptr;
    if( settings.value( FiltersUnrec, false ).toBool() ){
      compOr = new CompositeFilter();
      compOr->setOr();

      ReconcileFilter * yfilt = new ReconcileFilter( account );
      yfilt->setAntiReconciledState( Yes );
      compOr->add( std::unique_ptr<TransactionFilter>( yfilt ) );
    }

    if( 0 != months ){
      QDate date( QDate::currentDate() );
      QDate end;
      if( months > 0 ){
        date = date.addMonths( -months );
      }
      else if( months < 0 ){
        switch( months ){
          case -1:
            date.setYMD( date.year(), date.month(), 1 );
            end.setYMD( date.year(), date.month(), date.daysInMonth() );
            break;
          case -2:
            date = date.addMonths( -1 );
            date.setYMD( date.year(), date.month(), 1 );
            end = QDate::currentDate().addMonths( 1 );
            end.setYMD( end.year(), end.month(), end.daysInMonth() );
            break;
          case -12:
            date.setYMD( date.year(), 1, 1 );
            end.setYMD( date.year(), 12, 31 );
            break;
        }
      }

      DateRangeFilter * drf = new DateRangeFilter( date, end );
      std::unique_ptr<TransactionFilter> udrf( drf );
      if( nullptr == compOr ){
        viewer->addCustomFilter( std::move( udrf ) );
        filtering = true;
      }
      else {
        compOr->add( std::move( udrf ) );
        viewer->addCustomFilter( std::unique_ptr<TransactionFilter>( compOr ) );
        filtering = true;
      }
    }
  }
}

bool JournalViewer::isFiltering() const {
  return filtering;
}

void JournalViewer::changeAccount( const Account& a ) {
  for ( int i = 0; i < count( ); i++ ) {
    AccountViewer * viewer = static_cast<AccountViewer *> ( widget( i ) );
    viewer->changeAccount( a );
  }
}

void JournalViewer::sendEditTrans( const FullTransaction& ft ){
  editTrans( ft );
}

void JournalViewer::selectJournal( const Journal& j ) {
  setCurrentIndex( journalIdLkp[j.getId( )] );
}

void JournalViewer::saveSettings( ) {
  for ( int i = 0; i < count( ); i++ ) {
    AccountViewer * viewer = static_cast<AccountViewer *> ( widget( i ) );
    QString jname = QString( "JournalViewer/account%1state" ).arg( i );
    viewer->saveSettings( jname );
  }
}

void JournalViewer::restoreSettings( ) {
  for ( int i = 0; i < count( ); i++ ) {
    AccountViewer * viewer = static_cast<AccountViewer *> ( widget( i ) );
    QString jname = QString( "JournalViewer/account%1state" ).arg( i );
    viewer->restoreSettings( jname );
  }

  changeFilters();
}
