#include "LoanDialog.h"
#include "Loan.h"
#include "QHacc.h"
#include "LoanMapper.h"
#include "Logger.h"
#include "AccountMapper.h"
#include "AccountUtils.h"

#include <math.h>
#include <cmath>

LoanDialog::LoanDialog(QWidget *parent) : QDialog( parent ) {
  widget.setupUi( this );
  widget.entryarea->setEnabled( false );
  widget.deleter->setEnabled( false );

  connect( widget.newloan, SIGNAL( clicked( ) ), SLOT( makeLoan( ) ) );
  connect( widget.loanSelector,
      SIGNAL( currentItemChanged( QListWidgetItem *, QListWidgetItem * ) ),
      SLOT( loadLoan( QListWidgetItem * ) ) );

  QPushButton * btn = widget.editButtonBox->button( QDialogButtonBox::Save );
  connect( btn, SIGNAL( clicked( ) ), SLOT( update( ) ) );
  connect( widget.deleter, SIGNAL( clicked( ) ), SLOT( remove( ) ) );

  AccountMapper& amap = QHacc::getInstance( )->getAccountMapper( );
  AccountUtils& autils = QHacc::getInstance( )->getAccountUtils( );

  QHash<QString, int> principals;
  QHash<QString, int> interests;

  for( const Account& acct : amap.getAll( ) ) {
    const AccountType type = acct.getType();
    if( Liability == type ){
      principals.insert( autils.longName( acct ), acct.getId() );
    }
    else if( Expense == type ){
      interests.insert( autils.longName( acct ), acct.getId() );
    }
  }

  QStringList list( principals.keys() );
  list.sort();
  widget.mainacct->addItem( tr( "None Selected" ), IDableEntity::NOID );
  for( const QString& name : list ){
    widget.mainacct->addItem( name, principals[name] );
  }

  QStringList list2( interests.keys() );
  list2.sort();
  widget.intacct->addItem( tr( "None Selected" ), IDableEntity::NOID );
  for( const QString& name : list2 ){
    widget.intacct->addItem( name, interests[name] );
  }

  connect( widget.mainacct, SIGNAL( currentIndexChanged( int ) ),
           SLOT( checkIfSaveable() ) );
  connect( widget.intacct, SIGNAL( currentIndexChanged( int ) ),
           SLOT( checkIfSaveable() ) );

  refresh( );
}

void LoanDialog::makeLoan( ){
  Loan l;
  l.setName( tr( "New Loan" ) );
  LoanMapper& mapper = QHacc::getInstance()->getLoanMapper();
  mapper.save( l );

  QListWidgetItem * item = addItem( l );
  widget.loanSelector->setCurrentItem( item );
}

void LoanDialog::loadLoan( QListWidgetItem * idx ){
  widget.entryarea->setEnabled( 0 != idx );
  widget.deleter->setEnabled( 0 != idx );
  widget.mainacct->setCurrentIndex( widget.mainacct->findData( Account::NOID ) );
  widget.intacct->setCurrentIndex( widget.intacct->findData( Account::NOID ) );

  if ( 0 != idx ) {
    LoanMapper& mapper = QHacc::getInstance()->getLoanMapper();
    int tid = idx->data( Qt::UserRole ).toInt( );

    current = mapper.getById( tid );
    widget.intrate->setText( QString::number( current.getYearlyInterestRate() * 100 ) );
    widget.periods->setText(QString::number( current.getPeriods() ) );
    widget.namer->setText( current.getName() );
    widget.principal->setText( current.getAmount().toString() );

    widget.mainacct->setCurrentIndex(widget.mainacct->findData( current.getPrincipalAccountId() ) );
    widget.intacct->setCurrentIndex(widget.intacct->findData( current.getInterestAccountId() ) );

    updateLabels();
  }
}

void LoanDialog::update( ){
  current.setPeriods( widget.periods->text().toInt() );
  current.setName( widget.namer->text() );
  current.setAmount( Money::getMoney( widget.principal->text() ) );
  current.setYearlyInterestRate( widget.intrate->text().toDouble() / 100 );

  current.setPrincipalAccountId( widget.mainacct->itemData( widget.mainacct->currentIndex() ).toInt() );
  current.setInterestAccountId( widget.intacct->itemData( widget.intacct->currentIndex() ).toInt() );

  current.setPayment( Money::getMoney( widget.paymentlbl->text() ) );

  LoanMapper& mapper = QHacc::getInstance()->getLoanMapper();
  mapper.update( current );
  refresh();
}

void LoanDialog::remove( ){
  LoanMapper& mapper = QHacc::getInstance()->getLoanMapper();
  mapper.remove( current );
  refresh( );
}

void LoanDialog::refresh( ){
  widget.loanSelector->clear( );

  QString empty;
  widget.intrate->setText( empty );
  widget.namer->setText( empty );
  widget.principal->setText( empty );
  widget.periods->setText( empty );
  widget.interestlbl->setText( empty );
  widget.principallbl->setText( empty );
  widget.paymentlbl->setText( empty );

  for( const Loan& t : QHacc::getInstance( )->getLoanMapper().getAll() ) {
    addItem( t );
  }

  widget.deleter->setEnabled( false );
  widget.entryarea->setEnabled( false );
  checkIfSaveable();
}

QListWidgetItem * LoanDialog::addItem( const Loan& h ) {
  QListWidgetItem * item = new QListWidgetItem( widget.loanSelector );
  item->setText( h.getName( ) );
  item->setData( Qt::UserRole, h.getId() );
  return item;
}

void LoanDialog::updateLabels(){
  Money loanamt = Money::getMoney( widget.principal->text() );
  int loanamti = loanamt.getValue();
  double intrate = widget.intrate->text().toDouble() / 1200;
  int periods = widget.periods->text().toInt();

  double oneplusr = pow( 1 + intrate, periods );
  double payo = ceil( loanamti * ( intrate * oneplusr )/( oneplusr-1 ) );

  Money payment( static_cast<int>( payo ) );
  Money intment( static_cast<int>( ceil( intrate * loanamti ) ) );
  Money priment( payment - intment );

  if( payment.isZero() || intment.isZero() || priment.isZero() ){
    widget.interestlbl->setText( "" );
    widget.principallbl->setText( "" );
    widget.paymentlbl->setText( "" );
  }
  else {
    widget.interestlbl->setText( intment.toString() );
    widget.principallbl->setText( priment.toString() );
    widget.paymentlbl->setText( payment.toString() );
  }
}

void LoanDialog::checkIfSaveable(){
  QPushButton * btn = widget.editButtonBox->button( QDialogButtonBox::Save );

  btn->setEnabled(
        !( IDableEntity::NOID == widget.mainacct->itemData( widget.mainacct->currentIndex() ).toInt() ||
           IDableEntity::NOID == widget.intacct->itemData( widget.intacct->currentIndex() ).toInt() ) );
}
