/************************* * * * * * * * * * * * * ***************************
    Copyright (c) 1999-2013 Ryan Bobko
                       ryan@ostrich-emulators.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 ************************** * * * * * * * * * * * * **************************/

#include <QTreeWidget>
#include <QTreeWidgetItem>
#include <QSettings>
#include <QSignalMapper>
#include <QLabel>
#include <QButtonGroup>
#include <QPushButton>

#include "MainWindow.h"
#include "QHacc.h"
#include "Account.h"
#include "AccountMapper.h"
#include "Logger.h"
#include "GUIStore.h"
#include "PlotWindow.h"
#include "MonthlyBalanceWindow.h"
#include "MonthlyChangeWindow.h"
#include "DailyBalanceWindow.h"
#include "DailyChangeWindow.h"
#include "PieWindow.h"
#include "LoanDialog.h"
#include "MemTransDataEntry.h"
#include "AccountDialog.h"
#include "JournalMapper.h"
#include "ReconcileWindow.h"
#include "PreferencesDialog.h"
#include "FullTransaction.h"
#include "EngineNotifier.h"
#include "BudgetMapper.h"
#include "BudgetUtils.h"
#include "BudgetDialog.h"
#include "BudgetingWindow.h"
#include "QuickEntryDialog.h"
#include "AccountUtils.h"
#include "FilterViewDialog.h"

const int MainWindow::GRAPH_MBALANCE = 1;
const int MainWindow::GRAPH_MCHANGE = 2;
const int MainWindow::GRAPH_DBALANCE = 3;
const int MainWindow::GRAPH_DCHANGE = 4;
const int MainWindow::GRAPH_CPIE = 5;
const int MainWindow::GRAPH_DPIE = 6;
const int MainWindow::GRAPH_BUDGET = 7;

MainWindow::MainWindow( ){
  widget.setupUi( this );

  connect( GUIStore::getInstance( ),
           SIGNAL( transactionChanged( const FullTransaction & ) ),
           widget.dataentry, SLOT( show( ) ) );

  connect( GUIStore::getInstance( ),
           SIGNAL( transactionChanged( const FullTransaction & ) ),
           widget.dataentry, SLOT( edit( const FullTransaction& ) ) );

  connect( widget.jviewer, SIGNAL( journalSelected( const Journal & ) ),
           SLOT( changeJournal( const Journal& ) ) );

  connect( widget.achsr, SIGNAL( accountSelected( const Account& ) ),
           SLOT( changeAccount( const Account& ) ) );
  connect( widget.jviewer, SIGNAL( editTrans(const FullTransaction& ) ),
           SLOT( editTrans( const FullTransaction& ) ) );

  QSignalMapper * mapper = new QSignalMapper( this );
  connect( widget.actionMonthly_Balance, SIGNAL( activated() ),
           mapper, SLOT( map() ) );
  mapper->setMapping( widget.actionMonthly_Balance, GRAPH_MBALANCE );
  connect( widget.actionMonthly_Change, SIGNAL( activated() ),
           mapper, SLOT( map() ) );
  mapper->setMapping( widget.actionMonthly_Change, GRAPH_MCHANGE );
  connect( widget.actionDaily_Balance, SIGNAL( activated() ),
           mapper, SLOT( map() ) );
  mapper->setMapping( widget.actionDaily_Balance, GRAPH_DBALANCE );
  connect( widget.actionDaily_Change, SIGNAL( activated() ),
           mapper, SLOT( map() ) );
  mapper->setMapping( widget.actionDaily_Change, GRAPH_DCHANGE );
  connect( widget.actionExpense_Pie, SIGNAL( activated() ),
           mapper, SLOT( map() ) );
  mapper->setMapping( widget.actionExpense_Pie, GRAPH_CPIE );
  connect( widget.actionIncome_Pie, SIGNAL( activated() ),
           mapper, SLOT( map() ) );
  mapper->setMapping( widget.actionIncome_Pie, GRAPH_DPIE );
  connect( widget.actionBudget, SIGNAL( activated() ),
           mapper, SLOT( map() ) );
  mapper->setMapping( widget.actionBudget, GRAPH_BUDGET );

  connect( mapper, SIGNAL( mapped( int ) ), SLOT( openGraph( int ) ) );

  connect( widget.actionMemTrans, SIGNAL( activated( ) ),
           SLOT( openMemorizedTransactionEditor( ) ) );

  connect( widget.actionPreferences, SIGNAL( activated( ) ),
           SLOT( openPreferences() ) );

  connect( widget.actionCreate, SIGNAL( activated( ) ), SLOT( newAccount( ) ) );
  connect( widget.actionEdit, SIGNAL( activated( ) ), SLOT( editAccount( ) ) );
  connect( widget.actionReconcile, SIGNAL( activated( ) ),
           SLOT( openReconcile( ) ) );

  connect( widget.actionLoans, SIGNAL( activated() ), SLOT( openLoans() ) );
  connect( widget.actionBudgets, SIGNAL( activated() ), SLOT( openBudgets() ) );
  connect( widget.actionQuickEntry, SIGNAL( activated() ), SLOT( openQuickEntry() ) );

  connect( widget.actionAddToFaves, SIGNAL( activated() ),
           SLOT( addCurrentToFavorites() ));
  connect( widget.actionRemoveFromFaves, SIGNAL( activated() ),
           SLOT( removeCurrentFromFavorites() ));

  widget.achsr->refresh( );
  widget.jviewer->refresh( );
  widget.dataentry->hide( );

  // build the status bar
  messageText = new QLabel( tr( "Filtering Transactions" ) );
  messageText->setFrameStyle( QFrame::Panel | QFrame::Sunken );
  statusBar()->addPermanentWidget( messageText );
  balanceText = new QLabel();
  statusBar()->addPermanentWidget( balanceText );

  EngineNotifier * noter = QHacc::getInstance()->getNotifier();
  connect( noter, SIGNAL(accountChanged( const Account&, const Account& ) ),
           SLOT( accountUpdated( const Account&, const Account& ) ) );

  populars = new QButtonGroup( widget.popularaccounts );
  faves = new QButtonGroup( widget.favorites );
  connect( faves, SIGNAL( buttonClicked( int ) ),
           SLOT( quickAccountSelected( int ) ) );
  connect( populars, SIGNAL( buttonClicked( int ) ),
           SLOT( quickAccountSelected( int ) ) );

  initPopulars();
  initFromSettings( );
}

MainWindow::~MainWindow( ) {
  QHacc::cleanup( );
}

void MainWindow::editTrans(const FullTransaction& ft ){
  widget.dataentry->show();
  widget.dataentry->edit( ft );
}

void MainWindow::changeAccount( const Account& account ) {
  GUIStore::getInstance( )->setAccount( account );

  widget.newtrans->setEnabled( true );
  widget.actionMonthly_Balance->setEnabled( true );
  widget.actionMonthly_Change->setEnabled( true );
  widget.actionDaily_Balance->setEnabled( true );
  widget.actionDaily_Change->setEnabled( true );
  widget.actionAddToFaves->setEnabled( true );
  widget.actionRemoveFromFaves->setEnabled( true );
  widget.dataentry->hide( );
  widget.jviewer->changeAccount( account );

  accountUpdated( account, account );

  // save for next startup
  QSettings settings;
  settings.setValue( OpenAccountId, account.getId( ) );

  QAbstractButton * btn = faves->button( account.getId() );
  if( nullptr != btn ){
    btn->setChecked( true );
  }

  btn = populars->button( account.getId() );
  if( nullptr != btn ){
    btn->setChecked( true );
  }
}

void MainWindow::changeJournal( const Journal& journal ) {
  GUIStore::getInstance( )->setJournal( journal );
  widget.dataentry->changeJournal( journal );

  // save for next startup
  QSettings settings;
  settings.setValue( OpenJournalId, journal.getId( ) );
}

void MainWindow::accountUpdated( const Account&, const Account& account ){
  const Account& current = GUIStore::getInstance()->getAccount();
  if( account == current ) {
    QSettings settings;
    balanceText->setText( QString( "%1: %2%3 / Rec: %2%4" ).arg( account.getName() )
                          .arg( settings.value( CurrencySymbol ).toString() )
                          .arg( account.getBalance().toString() )
                          .arg( account.getBalance( Reconciled ).toString() ) );
  }
  checkBudgets( account );
}

void MainWindow::checkBudgets( const Account& acct ){
  BudgetMapper& bmap = QHacc::getInstance()->getBudgetMapper();
  BudgetUtils& butils = QHacc::getInstance()->getBudgetUtils();
  EngineNotifier * noter = QHacc::getInstance()->getNotifier();

  for( auto budget : bmap.getAllFor( acct.getId() ) ){
    if( butils.isOverBudget( budget ) ){
      noter->budget( budget );
    }
  }
}

void MainWindow::openGraph( int type ) {
  const Account& acct = GUIStore::getInstance( )->getAccount( );
  PlotWindow * pw = nullptr;
  switch( type ){
  case GRAPH_MBALANCE:
    pw = new MonthlyBalanceWindow( acct, this );
    break;
  case GRAPH_MCHANGE:
    pw = new MonthlyChangeWindow( acct, this );
    break;
  case GRAPH_DBALANCE:
    pw = new DailyBalanceWindow( acct, this );
    break;
  case GRAPH_DCHANGE:
    pw = new DailyChangeWindow( acct, this );
    break;
  case GRAPH_CPIE:
    pw = new PieWindow( acct, this );
    break;
  case GRAPH_DPIE:
    pw = new DebitPieWindow( acct, this );
    break;
  case GRAPH_BUDGET:
    pw = new BudgetingWindow( this );
    break;
  default:
    pw = new MonthlyBalanceWindow( acct, this );
  }

  pw->show( );
  pw->replot();
}

void MainWindow::openMemorizedTransactionEditor( ) {
  MemTransDataEntry * dlg = new MemTransDataEntry( this );
  dlg->show( );
}

void MainWindow::openReconcile( ) {
  const Account& a = GUIStore::getInstance( )->getAccount( );
  const Journal& j = GUIStore::getInstance( )->getJournal( );
  ReconcileWindow * dlg = new ReconcileWindow( a, j );
  dlg->show( );
}

void MainWindow::initFromSettings( ) {
  QSettings settings;
  if ( settings.contains( OpenAccountId ) ) {
    int aid = settings.value( OpenAccountId ).toInt( );
    AccountMapper& mapper = QHacc::getInstance( )->getAccountMapper( );
    Account account = mapper.getById( aid );
    widget.achsr->selectAccount( account );
  }
  if ( settings.contains( OpenJournalId ) ) {
    int jid = settings.value( OpenJournalId ).toInt( );
    JournalMapper& mapper = QHacc::getInstance( )->getJournalMapper( );
    Journal journal = mapper.getById( jid );
    widget.jviewer->selectJournal( journal );
  }

  restoreSettings( );

  if( settings.value( StartInQuickEntry, false ).toBool() ){
    openQuickEntry();
  }
}

void MainWindow::openViewFilters(){
  FilterViewDialog fvd;
  fvd.exec();
  QSettings settings;
  widget.jviewer->changeFilters();

  bool isfiltering = widget.jviewer->isFiltering();
  widget.filterbtn->setChecked( isfiltering );
  widget.filterlbl->setVisible( isfiltering );
  messageText->setVisible( isfiltering );
}

void MainWindow::newAccount( ) {
  AccountDialog dlg( Account( ), this );
  dlg.exec( );
}

void MainWindow::editAccount( ) {
  Account a = GUIStore::getInstance( )->getAccount( );
  AccountDialog dlg( a, this );
  dlg.exec( );
}

void MainWindow::saveSettings( ) {
  QSettings settings;
  settings.setValue( "MainWindow/state", saveState( ) );
  settings.setValue( "MainWindow/geometry", saveGeometry( ) );
  settings.setValue( "MainWindow/splitter", widget.splitter->saveState() );
  settings.setValue( "MainWindow/accounttab", widget.toolBox->currentIndex() );

  settings.beginWriteArray( "Favorites" );
  int count = 0;
  for( QAbstractButton * btn : faves->buttons() ){
    settings.setArrayIndex( count++ );
    settings.setValue( "account", faves->id( btn ) );
  }
  settings.endArray();

  widget.achsr->saveSettings( );
  widget.jviewer->saveSettings( );
}

void MainWindow::restoreSettings( ) {
  QSettings settings;
  restoreState( settings.value( "MainWindow/state" ).toByteArray( ) );
  restoreGeometry( settings.value( "MainWindow/geometry" ).toByteArray( ) );
  widget.splitter->restoreState( settings.value( "MainWindow/splitter").toByteArray() );
  widget.toolBox->setCurrentIndex( settings.value( "MainWindow/accounttab" ).toInt() );

  AccountMapper& amap = QHacc::getInstance( )->getAccountMapper( );
  int count = settings.beginReadArray( "Favorites" );
  for( int i = 0; i < count; i++ ){
    settings.setArrayIndex( i );
    int aid = settings.value( "account" ).toInt();
    try{
      Account acct = amap.getById( aid );
      modifyFavories( acct, true );
    }
    catch( MapperException me ){
      Logger::error( QString( "Unknown account: %1" ).arg( aid ) );
    }
  }
  settings.endArray();

  widget.achsr->restoreSettings( );
  widget.jviewer->restoreSettings( );
  bool isfiltering = widget.jviewer->isFiltering();
  widget.filterbtn->setChecked( isfiltering );
  widget.filterlbl->setVisible( isfiltering );
  messageText->setVisible( isfiltering );
}

void MainWindow::initPopulars(){
    AccountUtils& autils = QHacc::getInstance()->getAccountUtils();
    for( const Account& acct : autils.getMostPopularAccounts( Asset, 10 ) ){
      QPushButton * button = new QPushButton( acct.getName(), this );
      button->setCheckable( true );
      button->setSizePolicy( QSizePolicy::MinimumExpanding,
                           QSizePolicy::MinimumExpanding );
      populars->addButton( button, acct.getId() );
      widget.popularlayout->addWidget( button );
    }
}

void MainWindow::closeEvent( QCloseEvent* ) {
  saveSettings( );
}

void MainWindow::openPreferences(){
  PreferencesDialog pd( this );
  pd.exec();
}

void MainWindow::openLoans() {
  LoanDialog ld( this );
  ld.exec();
}

void MainWindow::openBudgets() {
  BudgetDialog ld( this );
  ld.exec();
}

void MainWindow::openQuickEntry(){
  QuickEntryDialog * qed = new QuickEntryDialog( this );
  qed->exec();
}

void MainWindow::quickAccountSelected( int aid ){
  AccountMapper& mapper = QHacc::getInstance( )->getAccountMapper( );
  Account account = mapper.getById( aid );
  widget.achsr->setAccount( account );
}

void MainWindow::addCurrentToFavorites(){
  const Account& acct = GUIStore::getInstance()->getAccount();
  modifyFavories( acct, true );
  QAbstractButton * btn = faves->button(acct.getId() );
  btn->setChecked( true );
}

void MainWindow::removeCurrentFromFavorites(){
  const Account& acct = GUIStore::getInstance()->getAccount();
  modifyFavories( acct, false );
}

void MainWindow::modifyFavories( const Account& acct, bool add ){
  if( add ){
    QPushButton * button = new QPushButton( acct.getName(), this );
    button->setCheckable( true );
    button->setSizePolicy( QSizePolicy::MinimumExpanding,
                           QSizePolicy::MinimumExpanding );
    faves->addButton( button, acct.getId() );
    widget.favoriteslayout->addWidget( button );
  }
  else {
    for( QAbstractButton * btn : faves->buttons() ){
      if( faves->id( btn ) == acct.getId() ){
        delete btn;
      }
    }
  }
}
