/*
 * File:   MemTransDataEntry.cpp
 * Author: ryan
 *
 * Created on November 27, 2011, 7:12 PM
 */

#include "MemTransDataEntry.h"
#include "QHacc.h"
#include "Transaction.h"
#include "TransactionUtils.h"
#include "TransactionMapper.h"
#include "SplitMapper.h"
#include "TransactionSchedule.h"
#include "Logger.h"
#include "AccountDelegate.h"
#include "ReconcileDelegate.h"
#include "JournalMapper.h"

#include <QListWidgetItem>
#include <QSettings>

MemTransDataEntry::MemTransDataEntry( QWidget * p ) : QDialog( p ) {
  widget.setupUi( this );
  setAttribute( Qt::WA_DeleteOnClose );

  widget.entryarea->setEnabled( false );
  widget.deleter->setEnabled( false );

  connect( widget.newtrans, SIGNAL( clicked( ) ), SLOT( makeTransaction( ) ) );
  connect( widget.transactionSelector,
      SIGNAL( currentItemChanged( QListWidgetItem *, QListWidgetItem * ) ),
      SLOT( loadTransaction( QListWidgetItem * ) ) );
  connect( widget.balancer, SIGNAL( clicked( ) ),
      widget.accounts, SLOT( addSplit( ) ) );
  connect( widget.saveBtn, SIGNAL( clicked( ) ), SLOT( update( ) ) );
  connect( widget.deleter, SIGNAL( clicked( ) ), SLOT( remove( ) ) );
  connect( widget.memo, SIGNAL( textChanged( const QString& ) ),
      widget.accounts, SLOT( changeMemo( const QString& ) ) );
  connect( widget.accounts, SIGNAL( differenceChanged( const Money&, const Money& ) ),
      this, SLOT( changedDifference( const Money&, const Money& ) ) );

  QSettings settings;
  QString fmt = settings.value( DateFormat ).toString();
  widget.nextrun->setDisplayFormat( fmt );

  refresh( );
}

void MemTransDataEntry::refresh( ) {
  widget.transactionSelector->clear( );

  widget.accounts->setSplits( QList<Split > ( ) );
  QString empty;
  widget.number->setText( empty );
  widget.payee->setText( empty );
  widget.memo->setText( empty );
  widget.transName->setText( empty );
  widget.shortcut->setText( empty );

  widget.journals->clear();
  JournalMapper& jmap = QHacc::getInstance()->getJournalMapper();
  for( const Journal& j : jmap.getAll() ){
    widget.journals->addItem( j.getName(), j.getId() );
  }

  TransactionMapper& tmap = QHacc::getInstance( )->getTransactionMapper( );
  QList<TransactionSchedule> trans = tmap.getSchedules();

  for( const TransactionSchedule& t : trans ) {
    addItem( t );
  }

  widget.deleter->setEnabled( false );
  widget.entryarea->setEnabled( false );
}

void MemTransDataEntry::makeTransaction( ) {
  FullTransaction ftrans;
  JournalMapper& jmap = QHacc::getInstance()->getJournalMapper();
  Transaction trans = ftrans.getTransaction();
  trans.setJournalId( jmap.getAll().at( 0 ).getId() );
  ftrans.reset( trans );

  TransactionSchedule handle( Unscheduled );
  handle.setName( tr( "New Memorized Transaction %1" ).
      arg( QDateTime::currentDateTime( ).toString( "MM/dd/yyyy hh:mm:ss.zzz" ) ) );
  QHacc::getInstance( )->getTransactionUtils( ).save( ftrans );
  QHacc::getInstance()->getTransactionMapper().schedule( ftrans.getTransaction(),
                                                         handle );
  QListWidgetItem * item = addItem( handle );
  widget.transactionSelector->setCurrentItem( item );
}

void MemTransDataEntry::loadTransaction( QListWidgetItem * idx ) {
  widget.entryarea->setEnabled( 0 != idx );
  widget.deleter->setEnabled( 0 != idx );

  if ( 0 != idx ) {
    int schedid = idx->data( Qt::UserRole ).toInt( );
    TransactionMapper& tmap = QHacc::getInstance()->getTransactionMapper();
    current = tmap.getSchedule( schedid );
    Transaction trans = tmap.getBySchedule( current );
    SplitMapper& smap = QHacc::getInstance()->getSplitMapper();
    QList<Split> splits = smap.getAll( trans );

    widget.accounts->setSplits( splits );
    widget.number->setText( trans.getNumber( ) );
    widget.payee->setText( trans.getPayee( ) );
    widget.transName->setText( current.getName( ) );

    int jid = trans.getJournalId();
    int jidx = widget.journals->findData( jid );
    widget.journals->setCurrentIndex( jidx );

    for( const Split& s : splits ) {
      widget.memo->setText( s.getMemo( ) );
    }

    setSchedule();
  }
}

QListWidgetItem * MemTransDataEntry::addItem( const TransactionSchedule & ts ) {
  QListWidgetItem * item = new QListWidgetItem( widget.transactionSelector );
  item->setText( ts.getName() );
  item->setData( Qt::UserRole, ts.getId() );
  return item;
}

void MemTransDataEntry::update( ) {
  TransactionMapper& tmap = QHacc::getInstance()->getTransactionMapper();
  TransactionUtils& tutils = QHacc::getInstance()->getTransactionUtils();

  QList<Split> splits( widget.accounts->getSplits( ) );
  Transaction trans = tmap.getBySchedule( current );
  trans.setNumber( widget.number->text( ) );
  trans.setPayee( widget.payee->text( ) );
  int jidx = widget.journals->itemData( widget.journals->currentIndex() ).toInt();
  trans.setJournalId( jidx );

  FullTransaction old = tutils.getTransaction( trans.getId() );
  FullTransaction ft( trans, splits );
  tutils.update( old, ft );

  current.setName( widget.transName->text( ) );
  ScheduleType type = Unscheduled;
  if( widget.rbannually->isChecked() ){
    type = Annually;
  }
  else if( widget.rbbimonthly->isChecked() ){
    type = BiMonthly;
  }
  else if( widget.rbbiweekly->isChecked() ){
    type = BiWeekly;
  }
  else if( widget.rblastday->isChecked() ){
    type = LastDay;
  }
  else if( widget.rbmonthly->isChecked() ){
    type = Monthly;
  }
  else if( widget.rbweekly->isChecked() ){
    type = Weekly;
  }
  else if( widget.rbdaily->isChecked() ){
    type = Daily;
  }

  current.setNextRun( widget.nextrun->date(), type );
  tmap.update( current );

  QListWidgetItem * item = widget.transactionSelector->currentItem( );
  item->setText( current.getName( ) );
}

void MemTransDataEntry::remove( ) {
  TransactionUtils& tutils = QHacc::getInstance( )->getTransactionUtils( );
  TransactionMapper& tmap = QHacc::getInstance( )->getTransactionMapper( );
  tutils.remove( tmap.getBySchedule( current ) );
  refresh( );
}

void MemTransDataEntry::changedDifference( const Money&, const Money& newer ) {
  widget.balancer->setText( QString( "%1: %2" ).arg( tr( "Add Split" ) ).
      arg( newer.toString( ) ) );
}

void MemTransDataEntry::setSchedule(){
  ScheduleType type = current.getType();
  QDate nexto( current.getNextRun() );
  if( !nexto.isValid() ){
    nexto = QDate( 1960, 1, 1 );
    type = Unscheduled;
  }
  widget.nextrun->setDate( nexto );

  widget.doschedule->setChecked( Unscheduled != type );
  QRadioButton * btn;
  switch( type ){
    case Monthly:
      btn = widget.rbmonthly;
      break;
    case BiWeekly:
      btn = widget.rbbiweekly;
      break;
    case BiMonthly:
      btn = widget.rbbimonthly;
      break;
    case Annually:
      btn = widget.rbannually;
      break;
    case LastDay:
      btn = widget.rblastday;
      break;
    case Weekly:
      btn = widget.rbweekly;
      break;
    case Daily:
      btn =widget.rbdaily;
      break;
    default:
      btn = widget.rbunsched;
  }

  btn->setChecked( true );
}

void MemTransDataEntry::setNextRunDate(){
  widget.nextrun->setDate( widget.rbunsched->isChecked() ? QDate( 1960, 1, 1 ) :
                                                           QDate::currentDate().addDays( -1 ) );
}
