/*
 * File:   PlotWindow.cpp
 * Author: ryan
 *
 * Created on November 21, 2011, 3:44 PM
 */

#include "PlotWindow.h"

#include <qwt_plot_curve.h>
#include <qwt_plot_grid.h>
#include <qwt_legend.h>

#include "PlotGenerator.h"
#include "MonthLabeler.h"
#include "DayLabeler.h"
#include "AccountLabeler.h"
#include "Logger.h"
#include "QHacc.h"
#include "AccountMapper.h"


PlotWindow::PlotWindow( PlotGenerator * pg, QWidget * p ) : QDialog( p ), loading( false ) {
  widget.setupUi( this );
  QDate today = QDate::currentDate( );
  widget.enddate->setDate( today );
  widget.startdate->setDate( today.addYears( -1 ) );
  setAttribute( Qt::WA_DeleteOnClose );
  connect( widget.plotButton, SIGNAL( clicked( ) ), SLOT( replot( ) ) );
  setGenerator( pg );

  connect( widget.addA, SIGNAL( clicked() ), SLOT( add() ) );
  init();
}

PlotWindow::~PlotWindow( ) {
  clearGenerators( );
}

void PlotWindow::init() {
}

void PlotWindow::replot( ) {
  if( loading ){
    return;
  }

  QDate starter = widget.startdate->date( );
  QDate ender = widget.enddate->date( );

  const TimeScale ts = timescale();
  QwtScaleDraw * drawer = nullptr;
  if( MONTHLY == ts ){
    drawer = new MonthLabeler( starter );
  }
  else if( DAILY == ts ){
    drawer = new DayLabeler( starter );
  }
  else if( ACCOUNTS == ts ){
    drawer = new AccountLabeler();
  }

  widget.qwtPlot->setAxisScaleDraw( QwtPlot::xBottom, drawer );

  widget.qwtPlot->detachItems();
  widget.qwtPlot->setAxisLabelRotation( QwtPlot::xBottom, -50.0 );
  widget.qwtPlot->setAxisLabelAlignment( QwtPlot::xBottom,
      Qt::AlignLeft | Qt::AlignBottom );

  const int gsize = generators.size( );
  QPen * pens = new QPen[gsize];
  if ( gsize > 1 ) {
    QStringList colornames = QColor::colorNames( );
    int csize = colornames.size( );

    for ( int i = 1; i < gsize; i++ ) {
      pens[i].setColor( colornames[( i - 1 ) % csize] );
    }
  }

  int i = 0;
  clearLegend();
  for( PlotGenerator * pg : generators ) {
    pg->setWindow( starter, ender );
    pens[i].setWidth( 3 );
    std::unique_ptr<QwtPlotItem> item = pg->getPlot( pens[i] );
    item.release()->attach( widget.qwtPlot );
    i++;

    for( const auto& lm : pg->legendMap() ){
      const std::pair<QColor, QString>& dato = lm.second;
      QLabel * lbl = new QLabel( dato.second );
      lbl->setAlignment( Qt::AlignCenter );

      bool isdark = ( dato.first == Qt::darkRed ||
                      dato.first == Qt::blue ||
                      dato.first == Qt::darkBlue ||
                      dato.first == Qt::darkMagenta ||
                      dato.first == Qt::darkGreen ||
                      dato.first == Qt::darkCyan ||
                      dato.first == Qt::darkYellow ||
                      dato.first == Qt::black );
      QColor foreg( isdark ? Qt::white : Qt::black );

      QString style = QString("QLabel{background-color: %1; color: %2; padding: 2px}")
          .arg( dato.first.name() ).arg( foreg.name() );
      lbl->setStyleSheet( style );
      lbl->setFixedWidth( 50 );
      widget.legendlayout->addRow( lm.first, lbl );
    }
  }

  if( NONE != ts ){
    QwtPlotGrid * grid = new QwtPlotGrid( );
    grid->enableY( true );
    grid->enableX( false );
    QPen pen( Qt::DotLine );
    pen.setColor( Qt::gray );
    grid->setPen( pen );
    grid->attach( widget.qwtPlot );
  }

  widget.qwtPlot->enableAxis( QwtPlot::yLeft , ( NONE != ts ) );
  widget.qwtPlot->enableAxis( QwtPlot::xBottom , ( NONE != ts ) );

  widget.qwtPlot->replot( );
  delete [] pens;
}

void PlotWindow::setGenerator( PlotGenerator * g ) {
  loading = true;
  clearGenerators( );
  addGenerator( g );
  loading = false;
}

void PlotWindow::addGenerator( PlotGenerator * g ) {
  generators.append( g );
  setWindowTitle( 1 == generators.size( ) ?
      g->getTitle() : tr( "Multiple Accounts" ) );
  replot( );
}

void PlotWindow::clearGenerators( ) {
  while ( !generators.isEmpty( ) ) {
    delete generators.takeFirst( );
  }
}

void PlotWindow::clearLegend(){
  QLayoutItem * item;
  if( widget.legendlayout->count() > 0 ){
    while( ( item = widget.legendlayout->takeAt( 0 ) ) != nullptr ) {
      delete item->widget();
      delete item;
    }
  }
}
