#include "QuickEntryDialog.h"

#include <QPushButton>
#include <QSettings>
#include <QApplication>

#include "QHacc.h"
#include "Account.h"
#include "AccountMapper.h"
#include "AccountUtils.h"
#include "TransactionUtils.h"
#include "Logger.h"

QuickEntryDialog::QuickEntryDialog( QWidget * parent ) : QDialog( parent ){
  widget.setupUi( this );

  AccountUtils& autils = QHacc::getInstance()->getAccountUtils();
  QSettings settings;
  QString df = settings.value( DateFormat ).toString();
  widget.dateEdit->setDisplayFormat( df );
  widget.dateEdit->setDate( QDate::currentDate() );

  QAbstractButton * btn = widget.buttonBox->button( QDialogButtonBox::Discard );
  btn->setText( tr( "Exit QHacc" ) );
  connect( btn, SIGNAL( clicked() ), SLOT( exitApp() ) );

  btn = widget.buttonBox->button( QDialogButtonBox::Close );
  btn->setText( tr( "Return to QHacc" ) );

  QVBoxLayout * vboxF = new QVBoxLayout();
  QVBoxLayout * vboxT = new QVBoxLayout();
  QVBoxLayout * vboxP = new QVBoxLayout();
  QVBoxLayout * vboxL = new QVBoxLayout();

  groupF = new QButtonGroup( this );
  groupT = new QButtonGroup( this );
  groupP = new QButtonGroup( this );

  groupF->setExclusive( true );
  groupT->setExclusive( true );
  groupP->setExclusive( true );

  connect( groupF, SIGNAL( buttonClicked( int ) ), SLOT( getToAccounts( int ) ) );
  connect( groupT, SIGNAL( buttonClicked( int ) ), SLOT( getPayees() ) );
  connect( groupP, SIGNAL( buttonClicked( int ) ), SLOT( moveToAmount() ) );
  connect( widget.amount, SIGNAL( textChanged( QString ) ),
           SLOT( checkSaveable() ) );

  for( const Account& acct : autils.getMostPopularAccounts( Asset, 5 ) ){
    QPushButton * button = new QPushButton( acct.getName(), this );
    button->setCheckable( true );
    button->setSizePolicy( QSizePolicy::MinimumExpanding,
                           QSizePolicy::MinimumExpanding );
    vboxF->addWidget( button );
    groupF->addButton( button, acct.getId() );
  }

  for( const Account& acct : autils.getMostPopularAccounts( Liability, 5 ) ){
    QPushButton * button = new QPushButton( acct.getName(), this );
    button->setCheckable( true );
    button->setSizePolicy( QSizePolicy::MinimumExpanding,
                           QSizePolicy::MinimumExpanding );
    vboxL->addWidget( button );
    groupF->addButton( button, acct.getId() );
  }

  widget.fromBox->setLayout( vboxF );
  widget.fromBox2->setLayout( vboxL );
  widget.toBox->setLayout( vboxT );
  widget.payeeBox->setLayout( vboxP );
}

void QuickEntryDialog::getToAccounts( int aid ){
  AccountUtils& autils = QHacc::getInstance()->getAccountUtils();
  QList<Account> pairs = autils.getMostPopularPairAccounts( aid, Expense, 10 );

    // clear out the old buttons
  QLayoutItem * item;
  while( ( item = widget.toBox->layout()->takeAt( 0 ) ) != nullptr ) {
    delete item->widget();
    delete item;
  }
  while( ( item = widget.payeeBox->layout()->takeAt( 0 ) ) != nullptr ) {
    delete item->widget();
    delete item;
  }

  // add new buttons for accounts
  for( const Account& a : pairs ){
    QPushButton * button = new QPushButton( a.getName(), this );
    button->setCheckable( true );
    button->setSizePolicy( QSizePolicy::MinimumExpanding,
                           QSizePolicy::MinimumExpanding );
    groupT->addButton( button, a.getId() );
    widget.toBox->layout()->addWidget( button );
  }
}

void QuickEntryDialog::getPayees(){
  QAbstractButton * fbtn = groupF->checkedButton();
  QAbstractButton * tbtn = groupT->checkedButton();
  if( 0 != fbtn && 0 != tbtn ){
    int faid = groupF->id( fbtn );
    int taid = groupT->id( tbtn );

    // clear out the old buttons
    QLayoutItem * item;
    while( ( item = widget.payeeBox->layout()->takeAt( 0 ) ) != nullptr ) {
      delete item->widget();
      delete item;
    }

    // add new buttons for payees
    TransactionUtils& tutils = QHacc::getInstance()->getTransactionUtils();
    for( const QString& payee : tutils.getMostPopularPayees( faid, taid, 10 ) ){
      QPushButton * button = new QPushButton( payee, this );
      button->setCheckable( true );
      button->setSizePolicy( QSizePolicy::MinimumExpanding,
                             QSizePolicy::MinimumExpanding );
      groupP->addButton( button );
      widget.payeeBox->layout()->addWidget( button );
    }
  }
}

void QuickEntryDialog::checkSaveable(){
  QAbstractButton * fbtn = groupF->checkedButton();
  QAbstractButton * tbtn = groupT->checkedButton();
  Money money = Money::getMoney( widget.amount->text() );
  widget.saveButton->setEnabled( 0 != fbtn && 0 != tbtn && money.isNonZero() );
}

void QuickEntryDialog::save() {
  QSettings settings;
  int journalid = settings.value( OpenJournalId ).toInt();
  QAbstractButton * btn = groupP->checkedButton();
  QString payee;
  Money money = Money::getMoney( widget.amount->text() );
  if( nullptr != btn ){
    payee = btn->text();
  }

  Transaction t;
  t.setJournalId( journalid );
  t.setDate( widget.dateEdit->date() );
  t.setPayee( payee );

  Split s1;
  btn = groupF->checkedButton();
  s1.setAccountId( groupF->id( btn ) );

  s1.setValue( money.opposite() );

  Split s2;
  btn = groupT->checkedButton();
  s2.setAccountId( groupT->id( btn ) );
  s2.setValue( money );

  QList<Split> splits;
  splits << s1 << s2;

  FullTransaction ft( t, splits );

  TransactionUtils& tutils = QHacc::getInstance()->getTransactionUtils();
  tutils.save( ft );

  widget.amount->setText( "" );
  widget.amount->setFocus();
}

void QuickEntryDialog::moveToAmount() {
  widget.amount->setFocus();
}

void QuickEntryDialog::exitApp() {
  exit( 0 );
}
