/*
 * File:   SplitModel.cpp
 * Author: ryan
 *
 * Created on November 27, 2011, 9:07 PM
 */

#include "SplitModel.h"
#include "QHacc.h"
#include "AccountMapper.h"
#include "GUIStore.h"

SplitModel::SplitModel( QObject * p ) : QAbstractTableModel( p ) {
  headers << tr( "Account" ) << tr( "Memo" ) << tr( "Credit" )
          << tr( "Debit" ) << tr( "Rec", "reconciled field in account viewer" );
}

void SplitModel::setSplits( QList<Split> s ) {
  difference = 0;
  splits = s;
  figureDifference( );
  reset( );

  if ( difference == 0 ) {
    // emit differenceChanged even if we don't have a balance.
    emit differenceChanged( difference, difference );
  }
}

int SplitModel::rowCount( const QModelIndex & parent ) const {
  return ( parent.isValid( ) ? 0 : splits.size( ) );
}

int SplitModel::columnCount( const QModelIndex & ) const {
  return headers.count( );
}

QVariant SplitModel::data( const QModelIndex &index, int role ) const {
  if ( !( index.isValid( ) &&
          ( Qt::DisplayRole == role || Qt::UserRole == role ) ) ) {
    return QVariant( );
  }

  const Split& split = splits[index.row( )];
  const int COL = index.column( );

  if ( Qt::UserRole == role ) {
    if ( 0 == COL ) {
      return split.getAccountId( );
    }
    else if ( 4 == COL ) {
      return split.getReconciled( );
    }
  }

  if ( Qt::DisplayRole == role ) {
    switch ( COL ) {
    case 0:
      return ( split.hasAccount( ) ?
                 QHacc::getInstance( )->getAccountMapper( ).
                 getById( split.getAccountId( ) ).getName( ) : QVariant( ) );
    case 1:
      return split.getMemo( );
    case 2:
    {
      Money money( split.getValue( ) );
      return ( money.isPositive( ) ? money.toPositiveString( ) : QVariant( ) );
    }
    case 3:
    {
      Money money( split.getValue( ) );

      return ( money.isPositive( ) ? QVariant( ) : money.toPositiveString( ) );
    }
    case 4:
    {
      const ReconcileState rs = split.getReconciled( );
      if ( Yes == rs ) {
        return tr( "R", "reconciled" );
      }
      else if ( Maybe == rs ) {
        return tr( "C", "cleared" );
      }
      else return "";
    }
    }
  }
  return QVariant( );
}

bool SplitModel::setData( const QModelIndex & idx, const QVariant & val,
                          int role ) {
  bool ok = false;

  if ( idx.isValid( ) && Qt::EditRole == role ) {
    const int ROW = idx.row( );
    switch ( idx.column( ) ) {
    case 0:
      splits[ROW].setAccountId( val.toInt( ) );
      ok = true;
      break;
    case 1:
      splits[ROW].setMemo( val.toString( ) );
      ok = true;
      break;
    case 2:
      splits[ROW].setValue( Money::getMoney( val.toString( ) ) );
      ok = true;
      figureDifference( );
      break;
    case 3:
    {
      Money money( Money::getMoney( val.toString( ) ) );
      if ( money.isPositive( ) ) {
        money.negate( );
      }
      splits[ROW].setValue( money );
      ok = true;
      figureDifference( );
    }
    break;
    case 4:
      splits[ROW].setReconciled( static_cast<ReconcileState> ( val.toInt( ) ) );
      ok = true;
    }
  }

  return ok;
}

QVariant SplitModel::headerData( int col, Qt::Orientation orient, int role ) const {
  if ( Qt::DisplayRole == role && Qt::Horizontal == orient ) {
    return headers[col];
  }
  return QVariant( );
}

Qt::ItemFlags SplitModel::flags( const QModelIndex & ) const {
  return ( Qt::ItemIsEditable | Qt::ItemIsSelectable | Qt::ItemIsEnabled );
}

Money SplitModel::getDifference( ) const {
  Money m;

  for( const Split& split : splits ) {
    m += split.getValue( );
  }

  return m;
}

void SplitModel::figureDifference( ) {
  Money old( difference );
  difference = getDifference( );

  if ( difference != old ) {
    emit differenceChanged( old, difference );
  }
}

void SplitModel::addSplit( ) {
  Split s;
  s.setMemo( defaultMemo );
  s.setValue( -difference );
  splits << s;
  reset( );
  figureDifference( );
}

const QList<Split>& SplitModel::getSplits( ) const {
  return splits;
}

void SplitModel::changeMemo( const QString& str ) {
  defaultMemo = str;
  for ( QList<Split>::iterator i = splits.begin( ); i != splits.end( ); ++i ) {
    i->setMemo( defaultMemo );
  }
  reset( );
}
