#include "PieGraph.h"
#include "Logger.h"
#include "QHacc.h"
#include "TransactionUtils.h"
#include "AccountUtils.h"
#include "AccountMapper.h"

#include <QPainter>

#include <qwt_scale_map.h>
#include <qwt_plot_curve.h>

PieGraph::PieGraph( const Account& a, bool creds, const QDate& b,
                    const QDate &e ) : AccountGenerator( a, b, e ),
                    credits( creds ) {
}

PieGraph::PieGraph( const PieGraph& orig ) : AccountGenerator( orig ),
  credits( orig.credits ){
}

PieGraph& PieGraph::operator=( const PieGraph& pg ) {
  if ( this != &pg ) {
    AccountGenerator::operator=( pg );
    credits = pg.credits;
  }

  return *this;
}

std::unique_ptr<QwtPlotItem> PieGraph::getPlot( const QPen& ) {
  QHacc * engine = QHacc::getInstance( );
  const AccountUtils& autils = engine->getAccountUtils( );
  QList<FullTransaction> trans =
      engine->getTransactionUtils( ).getTransactions( account, Journal( ),
      stop, start );

  std::map<int, Money> monies; // accountid, total money
  Money total;

  for( const FullTransaction& ft: trans ){
    // for each fulltrans, figure out if this account received or expended money
    // if it got money, figure out which account provided it. Hopefully, every
    // split trans has a single account providing all the money instead of
    // multiple credits and multiple debits in one fulltrans
    const Split& me = ft[account];
    const Money& memoney = me.getValue();

    bool include = ( ( !credits && memoney.isNegative() ) ||
                   ( credits && memoney.isPositive() ) );
    Logger::debug( QString( "checking %3 %1 => %2" ).arg( memoney.toString() ).arg( include ? "include" : "skip" ).arg( ft.getTransaction().getPayee() ) );

    if( include ){
      // include this trans in the calculations
      for( const Split& s : ft.getSplits() ){
        const int aid = s.getAccountId();
        Money m = s.getValue();
        if( memoney.oppositeSign( m ) ){
          if( m.abs() > memoney.abs() ){
            m = memoney;
          }

          m = m.abs();

          // FIXME: if we have a split with >1 negative accounts, we'll
          // mistakenly assign the same positive number to both accounts

          monies[aid] += m;
          total += m;
          Logger::debug( QString( "  accepting %1: %2" ).arg( autils.longName( aid ) ).arg( m.toString() ) );
        }
      }
    }
  }

  PieMarker * pm = new PieMarker( monies );
  std::unique_ptr<QwtPlotItem> pie( pm );

  legendInfo.clear();
  const std::map<QString, double>& pcts = pm->percentMap();
  const std::map<QString, QColor>& cols = pm->colorMap();
  // const Money& m = pm->total();
  for( const auto& p : pcts ){
    const QString& name = p.first;
    const double& pct = p.second;
    const QColor& color = cols.at( name );
    QString info = QString( "%1\%" ).arg( pct * 100, 0, 'f', 2 );

    auto dato = std::make_pair( color, info );
    legendInfo.insert( std::make_pair( name, dato ) );
  }
  return pie;
}

std::map<QString, std::pair<QColor, QString> > PieGraph::legendMap() const {
  return legendInfo;
}

PieMarker::PieMarker( const std::map<int, Money>& atots ) : mtotal( 0 ){
  setZ( 1000 );
  setRenderHint( QwtPlotItem::RenderAntialiased, true );

  static const unsigned int CSIZE = 14;
  static const Qt::GlobalColor COLORS[] = {
    Qt::green,
    Qt::blue,
    Qt::yellow,
    Qt::red,
    Qt::cyan,
    Qt::magenta,
    Qt::gray,
    Qt::darkRed,
    Qt::darkBlue,
    Qt::darkYellow,
    Qt::darkGreen,
    Qt::darkCyan,
    Qt::darkMagenta,
    Qt::lightGray };

  QHacc * engine = QHacc::getInstance( );
  const AccountUtils& autils = engine->getAccountUtils();
  AccountMapper& amap = engine->getAccountMapper();

  for( auto& am : atots ){
    mtotal += am.second;
  }

  for( auto& am : atots ){
    QString str = autils.longName( amap.getById( am.first ) );
    double pct = am.second / mtotal;
    accountTotals[str] = pct;
    Logger::debug( QString( "  account/pct %1=>%2" ).arg( str ).arg( pct ) );
  }

  if( mtotal.isZero() ){
    accountTotals["No transactions"] = 1.0;
  }

  if( atots.size() > CSIZE ){
    mergeSmallAccounts( CSIZE );
  }

  int coloridx = 0;
  for( auto& am : accountTotals ){
    accountColors[am.first] = QColor( COLORS[coloridx++] );
  }
}

int PieMarker::rtti() const {
    return QwtPlotItem::Rtti_PlotUserItem;
}

void PieMarker::draw( QPainter * p, const QwtScaleMap&, const QwtScaleMap& ymap,
    const QRectF &rect ) const {
  const int margin = 5;
  QRect pieRect;
  pieRect.setX( rect.x() + margin );
  pieRect.setY( rect.y() + margin );
  pieRect.setHeight( ymap.transform( 100.0 ) );
  pieRect.setWidth( pieRect.height() );

  int angle = static_cast<int>( 5760 * 0.75 );
  p->save();

  for( auto& am : accountTotals ){
    QString str = am.first;
    double pct = am.second;
    Logger::debug( QString( "account/bal %1=>%2" ).arg( str ).arg( pct ) );

    const int value = static_cast<int>( 5760 * pct );

    p->setBrush( QBrush( accountColors.at( am.first ), Qt::SolidPattern ) );
    p->drawPie( pieRect, -angle, -value );

    angle += value;
  }

  p->restore();
}

void PieMarker::mergeSmallAccounts( unsigned int maxpies ){
  // we have too many accounts, so merge the smallest ones until we have
  // enough colors for all of them. We'll do this in waves, to try to
  // merge the smallest accounts before bigger ones. It would probably
  // be smarter to just sort the values, and start gobbling them up from
  // the small end

  double PCTS[]={ 0.25, 0.5, 0.75, 1.0 };
  int PSIZE = 4;
  double otherpct = 0;
  for( int i = 0; i<PSIZE && accountTotals.size() > maxpies ; ++i ){
    const double MINPCT = PCTS[i] / maxpies;

    std::map<QString, double>::iterator it( accountTotals.begin() );
    while( it != accountTotals.end() ){
      if( it->second < MINPCT ){
        otherpct += it->second;
        accountTotals.erase( it++ );
      }
      else {
        it++;
      }
    }
  }

  QString str = QObject::tr( "Other", "Pie Graph \"Other\" slice" );
  accountTotals[str] = otherpct;
}

const std::map<QString, QColor>& PieMarker::colorMap() const {
  return accountColors;
}

const std::map<QString, double>& PieMarker::percentMap() const {
  return accountTotals;
}
